(* ::Package:: *)

(* ::Title:: *)
(*cellaut*)


(* ::Subsubtitle:: *)
(*author:		Klaus Sutner*)
(*affiliation:	Carnegie Mellon University*)
(*email:		sutner@cs.cmu.edu*)
(*url:         		http://www.cs.cmu.edu/~sutner/automata.html*)
(*version:  	11*)
(**)
(***PREPACKAGE***)


(* ::Section::Closed:: *)
(*Header*)


BeginPackage["Automata11`cellaut`"];
Begin["`Private`"];
End[];
EndPackage[];


(* ::Section::Closed:: *)
(*Usage*)


WidthCA::usage = "WidthCA[C] returns the width of cellular automtaton C.";
AlphabetCountCA::usage = "AlphabetCA[C] returns the cardinality of the alphabet of cellular automaton C.";
RuleCA::usage = "RuleCA[C] returns the rule of cellular automaton C.";
GenerateCA::usage = "GenerateCA[R,w,k,opts] generates a cellular automaton ow width w over alphabet k.";


(* ::Section:: *)
(*Cellular Automata*)


WidthCA[CA[w_,k_,r_]] := w;
AlphabetCountCA[CA[w_,k_,r_]] := k;
RuleCA[CA[w_,k_,r_]] := r;


(* ::Subsubsection::Closed:: *)
(*LocalMapCA*)


(* ::Input:: *)
(*ClearAll[LocalMapCA,domCodomCA]*)


Attributes[LocalMapCA] = {HoldFirst};
Options[LocalMapCA] = {Kary -> True, Full -> False};
LocalMapCA[ ff_, C_CA, opts : OptionsPattern[] ] := 
(
   Clear[ff];
   If[ ! OptionValue[Full],
          	AssignFunction[ ff, Sequence @@ domCodomCA[C], Kary ->  OptionValue[Kary] ],
    (* else *)
    	ff = Last[domCodomCA[C]]
    ]
);
domCodomCA[ C_CA  ] := 
      With[ {r = RuleCA[C], w   = WidthCA[C], k = AlphabetCountCA[C]},
   	{ IntegerDigits[Range[0, k^w - 1], k, w], Reverse[IntegerDigits[r, k, k^w]] } ];


(* ::Subsubsection::Closed:: *)
(*GlobalMapCA*)


Attributes[GlobalMapCA]={HoldFirst};
Options[GlobalMapCA]={Boundary->Cyclic,BoundaryZero->0};
GlobalMapCA[ ff_, C_CA, opts:OptionsPattern[] ] := 
Module[ {w,w2,loc},
    ClearAll[ff];
    w   = WidthCA[C];
    w2  = Ceiling[w/2];
    LocalMapCA[loc,C,Kary->False];
    ff = Switch[ OptionValue[Boundary],
     	    Cyclic, loc/@Partition[#,w,1,w2]&,
     	    Fixed,  loc /@ Partition[#,w,1,w2,OptionValue[BoundaryZero]]&,
     	    None,   loc /@ Partition[#,w,1]&
         ];
];


(* ::Subsubsection::Closed:: *)
(*GenerateCA*)


Options[GenerateCA]={Type->Function};GenerateCA[ ff_, w_Integer?Positive, k_Integer?Positive, opts:OptionsPattern[] ] := 
Module[ {A,r},
	A = Tuples[ Range[0,k-1], w ];
	r = Switch[ OptionValue[Type],
		Function, ff @@@ A,
		Rule,          A /. ff,
		Boole,       Boole[ ff @@@ (A /. {0->False,1->True}) ]
	];
	CA[w,k,FromDigits[ Reverse[r], k ]]
];


(* ::Subsubsection::Closed:: *)
(*ToFredkinCA*)


ToFredkinCA[ C_CA ] := 
Module[ {f, ff, k = AlphabetCountCA[C], w = WidthCA[C]},
    	LocalMapCA[ f, C ];
    	ff[xx__] := With[ { zz = Mod[{xx}, k ], ci = Ceiling[w/2] },
			  k zz[[ci]] + Mod[ f @@ zz + Quotient[{xx}[[ci]], k], k] ];
    	GenerateCA[ff, w, k^2]
];


(* ::Subsubsection::Closed:: *)
(*ToWidthTwoCA*)


ToWidthTwoCA[ C_CA ] := 
Module[ {w, k, ff},
	w = WidthCA[C];
	k = AlphabetCountCA[C];
	GlobalMapCA[ff, C, Boundary -> None];
	CA[ 2, k^(w - 1), 
	    FromDigits[Reverse[FromDigits[#,k] & /@ ff /@ (Tuples[k,2w-2]-1)], 
		       k^(w - 1)] ]
];


(* ::Subsubsection::Closed:: *)
(*ComposeCA*)


ClearAll[ComposeCA]
Options[ComposeCA]={};
ComposeCA[ C1_CA, C2_CA, opts:OptionsPattern[] ] :=
    Module[ {k,alph,w,f1,f2,B,norm},
    w = WidthCA[C1] + WidthCA[C2] - 1;
   k = AlphabetCountCA[C1];
    GlobalMapCA[f1,C1,Boundary-> None];
    LocalMapCA[f2,C2, Kary->False];
    B = f2[f1[#]]& /@ Reverse[Tuples[Range[0,k-1],w]];Echo[B];
    CA[ w,k, FromDigits[B,k] ]
        ]  /;  AlphabetCountCA[C1]===AlphabetCountCA[C2];
ComposeCA[C1_CA,C2_CA,C__CA,opts:OptionsPattern[]] := 
         ComposeCA[ComposeCA[C1,C2,opts],C,opts];
ComposeCA[C_CA,opts:OptionsPattern[]] := C;



(* ::Subsubsection::Closed:: *)
(*OrbitCA*)


Options[OrbitCA]={Boundary-> Cyclic,Full-> False};
OrbitCA[ C_CA, pat_List, t_Integer:40, opts:OptionsPattern[] ] := 
Module[ {rho},
        GlobalMapCA[  rho, C,SelectOptions[GlobalMapCA,opts] ];
        NestList[ rho, pat, t ]
      ];


(* ::Subsubsection::Closed:: *)
(*ToRulesCA*)


Options[ToRulesCA]={};
ToRulesCA[C_CA,opts:OptionsPattern[]]:=
Module[{k,w,n,r},
	r=RuleCA[C];
	w=WidthCA[C];
	k=AlphabetCountCA[C];
	n=k^w;
	Thread[Tuples[Range[0,k-1],w]->Reverse@IntegerDigits[r,k,n]]
];


(* ::Subsubsection::Closed:: *)
(*PrintCA*)


PrintCA[C_CA] := 
    Module[ {rules,k},
	rules = ToRulesCA[C];
	k = AlphabetCountCA[C];
	Assert[ k<10,"Alphabet too large."];
	rules = MapAt[ToWord[#+1,Alpha[k,Type->Digits]]&, rules,{All,1}];
          Switch[ WidthCA[C],
            	1,       {rules},
            	2|3,  Partition[ rules, AlphabetCountCA[C]^(WidthCA[C]-1) ],
            	_,        Partition[ rules, AlphabetCountCA[C]^(WidthCA[C]-2) ] 
            ]
          ] // Thread //  TableForm[#,TableSpacing->{1,2}]&;


(* ::Subsubsection::Closed:: *)
(*ToFACA*)


ToFACA[C_CA,opts:OptionsPattern[]]:=
Module[{n,Rn,k,K,w,lab,B1,B2,norm,tr,res},
	w=WidthCA[C];
	k=AlphabetCountCA[C];
	n=k^(w-1);
	lab=IntegerDigits[RuleCA[C],k,k^w]+1;
	B1=Flatten@Thread@Table[Range[n],{k}];
	B2=Join@@Table[Range[n],{k}];
	tr=Pairs[B1,Reverse@lab,B2];
	FA[ TSys[ n, k, k^w, tr, Alpha[k,Type->Digit]], All, All ]
];


(* ::Subsubsection::Closed:: *)
(*ClassifyCA*)


$classcamsg={"none","surjective","open","injective"};
Options[ClassifyCA]={Full->False};
ClassifyCA[C_CA,opts:OptionsPattern[]]:=
Module[{full,n,sa,sa2,G,H,comps,ntcind,ntscc,HH,diac,res},
	full = OptionValue[Full];
	If[Length[Union[Last/@Frequencies[LocalMapCA[k,C,Full->True]]]]>1,
		Return[0]];
	n = AlphabetCountCA[C]^(WidthCA[C]-1);
	sa = ToFACA[C];
	sa2 = ProductFA[ sa, sa];
	G = TransitionsToGraph[sa2];
	{H,comps,ntcind} = CondensationGraph[G,Full->True,GraphLayout->None];
	ntscc = comps[[ntcind]];
	HH = Subgraph[H,Intersection[ Union[FlattenOne[VertexOutComponent[H,#]& /@ ntscc]],
	Union[FlattenOne[VertexOutComponent[ReverseGraph[H],#]& /@ ntscc]] ]];
	diac = ntscc[[PositionOne[ntscc,1]]];
	res = If[ Length[diac]=!=n, 
		0, 
		Which[ 
			Length[ntscc]==1, 3,
			VertexDegree[HH,diac]==0, 2,
			True, 1 ]
	];
	If[ OptionValue[Full], 
		{res,G,H,comps,ntcind},
		res
	]
];


(* ::Subsubsection::Closed:: *)
(*InverseCA*)


Options[InverseCA]={Full->False};
InverseCA[CA[1,k_Integer,r_Integer]]:=
Module[{lab},
	LocalMapCA[lab, CA[1,k,r], Full->True ];
	Assert[ PermutationListQ[lab+1]];
	CA[ 1, k, FromDigits[ Reverse[InversePermutation[lab+1]-1], k ] ]
];


InverseCA[C_CA,opts:OptionsPattern[]]:=
Module[{Wl,Wr,defl,defr,ww,sa,Lr,Ll,join,LL,k,alph,ll,rul,lab,res},
	{Wl,Wr}=Last[WelchCA[C,Full->True]];
	defl=DefiniteQFA[Wl,Full->True];
	defr=DefiniteQFA[Wr,Full->True];
	ww=defl+defr;
	sa=ToSA[C];
	Lr=auxFullanguage[sa,defl];
	Ll=auxFullanguage[ReverseFA[sa],defr];
	join[{L1_List,L2_List}]:=Apply[StringJoin,CartesianProduct[StringReverse/@L1,L2],1];
	LL=Sort[FlattenOne[MapIndexed[CartesianProduct,join/@Pairs[Ll,Lr]]]];
	{k,alph}=Alphabet[C,Full->True];
	ll=k^(WidthCA[C]-1);
	rul=Thread[Range[ll]->Flatten[Thread[Table[alph,{ll/k}]]]];
	lab=LabelToCodeCA[Reverse[Last/@LL/. rul],ww,alph];
	res=CA[ww,ShortAlphabetSize[C],lab,ShortAlphabet[C],{}];
	res
];

auxFullanguage[M_FA,r_Integer]:=
Module[{n,tr,targ,extend},
	n=StateCount[M];
	tr=TransitionList[M];
	targ[p_Integer] := targ[p] = Rest /@ Cases[tr,{p,_,_} ] ;
	extend[{p_Integer,L_List,q_Integer}] := 
		Table[ {p,Append[L,x[[1]]],x[[2]]}, {x,targ[q]} ];
	Union[Most /@ Nest[ FlattenOne[extend/@#]&, Table[ {p,{},p}, {p,n} ], r ]]
];


(* ::Subsubsection::Closed:: *)
(*WelchCA*)


Options[ToWelchCA]={Full->True};
ToWelchCA[C_CA,opts:OptionsPattern[]]:=
Module[{sa,mr,mxr,pos,Wr,ml,mxl,Wl,nwm},
	sa = ToFACA[C];
	mr = ToKernelFA[ sa,{1},Full->True];
	mxr = Max[ Length/@ StateList[mr] ];
	pos=PositionList[StateList[mr],Select[StateList[mr],Length[#]==mxr&]];
	Wr = auxToDFA[SubFA[ mr, pos ]];
	ml = ToKernelFA[ ReverseFA[sa],{1},Full->True];
	mxl = Max[ Length/@ StateList[ml] ];
	pos=PositionList[StateList[ml],Select[StateList[ml],Length[#]==mxl&]];
	Wl = auxToDFA[SubFA[ ml, pos ]];
	nwm = AlphabetCountCA[C]^(WidthCA[C]-1)/(mxl mxr);
	If[ OptionValue[Full],
		{{mxl,nwm,mxr},{Wl,Wr}},
		{mxl,nwm,mxr}
	]
];
auxToDFA[ M_FA ] := 
DFA[ TSys[ StateCount[M], AlphabetCount[M], StateCount[M] AlphabetCount[M],
TransitionMatrix[M] /. {{x_Integer?Positive}:>x}, GetAlphabet[M],
 States->StateList[M],Type->Matrix ], 1, All ]



(* ::Subsubsection::Closed:: *)
(*SeedConfiguration*)


Options[SeedConfiguration]={Where->Center,Padding->0,Random->0};
SeedConfiguration[n_Integer?Positive,opts:OptionsPattern[]]:=
SeedConfiguration[{1},n,opts];
SeedConfiguration[s_Integer?NonNegative,n_Integer?Positive,opts:OptionsPattern[]]:=
SeedConfiguration[{s},n,opts];
SeedConfiguration[s_List,n_Integer?Positive,opts:OptionsPattern[]]:=
Module[{rnd,whr,pad,typ},
	If[0<(rnd=OptionValue[Random]),
		Return[RandomInteger[{0,rnd-1},n]]];
	whr=OptionValue[Where];
	pad=OptionValue[Padding];
	Flatten[Switch[whr,
			Center,PadLeft[s,n,pad,Floor[1/2 (n-Length[s])]],
			Right,PadLeft[s,n,pad],
			_,PadRight[s,n,pad]]
	]
];


(* ::Section::Closed:: *)
(*Footer*)


Null
