(* ::Package:: *)

(* ::Title:: *)
(*Experimental*)


(* ::Text:: *)
(*Author:		Klaus Sutner*)
(*Affiliation: 	Carnegie Mellon University*)
(*email:		sutner@cs.cmu.edu*)
(*url:			http://www.cs.cmu.edu/~sutner*)
(*version:     	10.4*)


(* ::Section::Closed:: *)
(*Header*)


BeginPackage["Automata11`experimental`"];
Begin["`Private`"];
End[];
EndPackage[];


(* ::Section::Closed:: *)
(*Formatting*)


(* ::Text:: *)
(*This needs major work. *)


$postrules={""->"\[Epsilon]",Eps->"\[Epsilon]","Eps"->"\[Epsilon]",Empty->\[EmptySet]};
$Post = ReplaceRepeated[#,$postrules]&;

MakeAbbrevs::msgon = "Abbreviations turned on, use MakeAbbrevs[False] to turn off.";
MakeAbbrevs[b_:True] := 
(
    LFA = LanguageFA;
    TF  = TableForm;
    TF2 = TableForm[#,TableDepth->2]&;
    TF1 = TableForm[#,TableSpacing->{1,1}]&;
    CF  = ColumnForm;
    MF  = MatrixForm;
    CP  = CartesianProduct;
    Message[MakeAbbrevs::msgon];
);

MakeAbbrevs[False|None] := 
Clear[LFA, TF, TF1, TF2, CF, MF, CP];


Clear[PrintTransitions]
PrintTransitions[K_Alpha][{p_,a_Integer,q_}] := 
	With[{KK = AlphabetList[K]}, Labeled[ p -> q, KK[[a]],Top ]];
PrintTransitions[K_Alpha][L:{{_,_Integer,_}...}] := 
	With[{KK = AlphabetList[K]}, Labeled[ #[[1]] -> #[[3]], KK[[#[[2]]]],Top ]& /@ L];


ClearAll[FormatAutomata];

FormatAutomata[Full]:=
Module[{prt},
    prt=Unprotect[FA,DFA,TD];
    ClearAll[FA,DFA,TD];
    Protect/@prt;
    $Post=.;
];

FormatAutomata[SizeOnly]:=
Module[{prt},
prt=Unprotect[FA,DFA,TD];
ClearAll[FA,DFA,TD];
FA/:MakeBoxes[FA[TSys[n_,k_,t_,__],__],ft_]:=
RowBox[{"FA[",MakeBoxes[n,ft],",",MakeBoxes[k,ft],",",MakeBoxes[t,ft],"]"}];
DFA/:MakeBoxes[DFA[TSys[n_,k_,t_,__],__],ft_]:=
RowBox[{"DFA[",MakeBoxes[n,ft],",",MakeBoxes[k,ft],",",MakeBoxes[t,ft],"]"}];
TD/:MakeBoxes[TD[TSys[n_,k_,t_,__],__],ft_]:=
RowBox[{"TD[",MakeBoxes[n,ft],",",MakeBoxes[k,ft],",",MakeBoxes[t,ft],"]"}];
Protect/@prt;
$Post=#1/.$postrules&;
];

FormatAutomata[None]:=
Module[{prt},prt=Unprotect[FA,DFA,TD];
ClearAll[FA,DFA,TD];
FA/:MakeBoxes[FA[__],_]:="-FA-";
DFA/:MakeBoxes[DFA[__],_]:="-DFA-";
TD/:MakeBoxes[TD[__],_]:="-TD-";
Protect/@prt;
$Post=#1/.$postrules&;];


EnumForm[L_List] := TableForm[L, TableHeadings -> {Automatic, None}];
EnumFormZero[L_List] := TableForm[L, TableHeadings -> {Range[0,Length[L]-1], None}];


$GraphMethods = {"SpringElectricalEmbedding", "SpringEmbedding", "LayeredDrawing", 
"LayeredDigraphDrawing", "RadialDrawing", "HighDimensionalEmbedding", 
"CircularEmbedding", "SpiralEmbedding", "LinearEmbedding", "RandomEmbedding"};


(* ::Text:: *)
(*This should use StringTemplate.*)


ClearAll[texExport]
texExport[X_] := 
Module[ {cmd},
        cmd = "emacs -batch /tmp/mtatexexport -l " <> $HomeDirectory <> 
        "/elisp/custom/process-mta.el -f save-buffer";
        Export["/tmp/mtatexexport", TraditionalForm[HoldForm[X]], "TeX"];
        Run[ cmd ]
];


(* ::Section::Closed:: *)
(*Combinatorics*)


NextSubset[ L_List, n_Integer ] := 
Module[ {i},
        i = -1;
        While [ L[[i]] == n + i + 1, i-- ]; 
        Join[  Drop[ L, i ], Range[-i] + L[[i]] ]
] /;  Length[L] <= n && L =!= Range[ n - Length[L] + 1, n ];

NextSubset[ L_List, n_Integer ] :=  Range[Length[L]]   /;  Length[L] <= n ;

NextPermutation[ LL_List ] := 
Module[ {a,b,L,n,i,j,jj},

	If[ LL==Reverse@Sort[LL], Return[Reverse[LL]] ];

	n = Length[LL];
	i = -2;
	While[ LL[[i]] >= LL[[i+1]], i-- ];
	
	a = LL[[i]];
	j = jj = i+1;
	b = LL[[jj]];
	While[ jj <= -1, 
		If[ b >= LL[[jj]] > a, b = LL[[jj]]; j = jj ];
		jj++
	];
	
	(* flip a and b, sort second part of list *)
	L = ReplacePart[ ReplacePart[LL,a,j],b,i];
	Join[ Drop[L,i+1], Reverse[Take[L,i+1]] ]
];

NextPermutation[ LL_List, n_Integer ] := NestList[ NextPermutation, LL, n ];


(* ::Text:: *)
(*Polya-Redfield counting.  Kludge.*)


Options[SubstitutePolynomial] = {Full -> False}; 
SubstitutePolynomial[pol_, n_Integer, m_Integer|m_Symbol, (opts___)?OptionQ] := 
If[ Full /. {opts} /. Options[SubstitutePolynomial], 
    pol /. Table[Subscript[x, i] -> Total@Array[Subscript[c, #1]^i& , {m}], {i, n}],
    pol /. Table[Subscript[x, i] -> m, {i, n}] 
]; 


(* ::Text:: *)
(*Do we really need this stuff? *)


prot=Unprotect[Xor];
Xor[L_List,K_List] := Union[Complement[L,K],Complement[K,L]];
Protect[Evaluate[prot]];

ClearAll[ListMultiply];
ListMultiply[ K_List, 0 ] :=  K;
ListMultiply[ K_List, 1 ] :=  1 - K;
ListMultiply[ K_List, L_List ] := Join@@( ListMultiply[K,#]& /@ L );

ListLessEqual[ L_List, K_List ] := And@@Thread[ L <= K ]   /; Length[L]==Length[K];
ListLess[ L_List, K_List ] := ListLessEqual[ L, K ] && L =!=K   /; Length[L]==Length[K];

MatrixLessEqual[ L_?MatrixQ, K_?MatrixQ ] := 
(L <= K //. LessEqual -> ListLessEqual)   /;  Dimensions[L] == Dimensions[K];
MatrixLess[ L_?MatrixQ, K_?MatrixQ ] := 
(L <= K //. Less -> ListLess)   /;  Dimensions[L] == Dimensions[K];


(* ::Section:: *)
(*Graphics*)


(* ::Section::Closed:: *)
(*Algebra*)


ClearAll[PadicOrder, PadicValuation];

PadicOrder[0, _] = Infinity;
PadicOrder[L_List, p_Integer?(#>1&)] := PadicOrder[#, p] & /@ L;
PadicOrder[x_Integer, p_Integer?(#>1&)] := IntegerExponent[x,p];
PadicOrder[x_Rational, p_Integer?(#>1&)] := 
    	PadicOrder[Numerator[x], p] - PadicOrder[Denominator[x], p];

PadicValuation[0, _] = 0;
PadicValuation[x_Integer | x_Rational, p_Integer?(#>1&)] := p^(-PadicOrder[x, p]);

GeneralizedMultiplicativeOrder[B_, m_] := 
		{0, MultiplicativeOrder[B, m]}    /; GCD[m, B] == 1;

GeneralizedMultiplicativeOrder[B_, m_] := 
Module[{prm, pom, poB},
	prm = First /@ FactorInteger[GCD[ B, m ]];
	pom = PadicOrder[m, #] & /@ prm;
	poB = PadicOrder[B, #] & /@ prm;
	{ Max[Ceiling[pom/poB]], MultiplicativeOrder[B, m/Times @@ (prm^pom)]}
];


Attributes[WitnessFunction]={HoldFirst};
WitnessFunction[ gg_, S_SG, W_List ] := 
		 WitnessFunction[ gg, ToList[S], W ] /; Length[S]==Length[W];
WitnessFunction[ gg_, S_List, W_List ] := 
(
	AssignFunction[ gg, S, W, Opts->Listable ];
	gg[DC[xx_]] := DC[gg[xx]]
);


(* ::Section::Closed:: *)
(*FSM*)


LassoDFA[t_Integer, p_Integer, F_List] := 
    With[ {n = t + p},
      DFA[ n, 1, {Join[ Range[2, n], {t + 1}]}, 1, F ] ];

LassoDFA[t_Integer, p_Integer, q_Integer] := LassoDFA[t, p, {q}];

AllLassoDFA[  t_Integer, p_Integer ] := LassoDFA[t, p, #] & /@ PowerSet[Range[t+p]];

AllLassoDFA[  n_Integer ] := FlattenOne[Table[AllLassoDFA[i,n-i],{i,0,n-1}]] ;


ToTransducerCA[ C_CA ] := 
With[ {k = AlphabetSize[C],w=WidthCA[C],
kk = Automata7`automata`Private`ShortAlphabetSize[C]},

	TD[ k^w, kk, kk, 
	Thread@Insert[Thread@Transitions@ToSA[C],Flatten@Table[i,{k^(w-1)},{i,k}],2],
	Range[k^w], Range[k^w] ] 
];

splitTrans[K1_, K2_][{p_,u_,v_,pp_}, n_] := 
Module[ {uu,vv,lu,lv,ll},
	uu = ToIndex[u,K1];
	vv = ToIndex[v,K2];
	lu = Length[uu];
	lv = Length[vv];
	ll = Max[lu,lv,1];
	uu = Join[uu,Table[0,{ll-lu}]];
	vv = Join[vv,Table[0,{ll-lv}]];
	{ Pairs[   Join[{p}, Range[n+1,n+ll-1]], uu, vv, Join[ Range[n+1,n+ll-1],{pp}] ],
	  n + ll - 1 }
];

SplitTransformations[ M_TD ] :=
Module[ {ttt,tt,t,nn,K1,K2},
	nn = First[M];
	K1 = Zwei[M];
	K2 = Drei[M];
	tt = M[[4]];
	ttt = {};
	Do[
		{t,nn} = splitTrans[K1,K2][ tt[[i]], nn ];
		ttt = Join[ ttt, t ],
		{i,Length[tt]}
	];
	ReplaceParts[ M, nn, 1, ttt, 4 ]
];


(*ZigzagDecompositionAGraph::usage = "ZigzagDecompositionAGraph[G] computes the zigzag decomposition of a
transitive (2,2)-regular graph G.";

ZigzagDecompositionAGraph[G_AGraph] :=
Module[{adj, adjr, zig, zag, makezz, ed, e1, e2, Z, Z1},
      adj = Edges[ToAdjacencyList[G]];
      adjr = Edges[ToAdjacencyList[ReverseAGraph[G]]];

      zig[{p_, q_}] :=
        If[adj[[p]] === {q, q}, 
		    {p, q}, 
		    {p, First[Complement[adj[[p]], {q}]]}];
      zag[{p_, q_}] :=
        If[adjr[[q]] === {p, p}, 
		     {p, q}, 
		     {First[Complement[adjr[[q]], {p}]], q}];
      makezz[{p_, q_}] :=
      (
          	Z1 = {{p, q}, zag[{p, q}]};
          	While[Length[Z1] == 2 || Take[Z1, 2] =!= Take[Z1, -2],
            		e1 = zig[Last[Z1]];
            		e2 = zag[e1];
            		Z1 = Join[Z1, {e1}, {e2}];
            	];
          	Drop[Z1, -2]
      );

      ed = Edges[ToEdgeList[G]];
      Reap[
          While[ ed =!= {},
            	Z = makezz[First[ed]];
            	Sow[Z];
            	ed = Complement[ed, Z];
      ]][[2, 1]]
];*)


Options[MarkPeriod]={Marker->{0->2,1->5}};
MarkPeriod[L_List,opts___?OptionQ]:=
Module[{qq,pp,mm,K},
  	qq = Flatten/@Cases[Position[L,#]&/@Union[L],{_,__}];
  	If[ qq==={}, Return[L] ];
  	pp = First@Sort[First/@qq];
  	K = L[[pp]] /. (Marker /. {opts} /. Options[MarkPeriod]);
  	L /. L[[pp]]->K
];


ClearAll[ToVaucanson]
ToVaucanson[M_?FSMQ, fl_String, opts:OptionsPattern[] ] :=
  
  Module[{n, k, alph, vlab, stlab, tr, st, ostrm, trf, writeStates, 
      writeTrans,statecoord},
    
    statecoord[p_] := 
      If[ OptionValue[Full], 
        StringJoin["(", "0", ",", ToString[p], ")"],
        StringJoin["(,)"]
        ];
    
    writeStates[p_] := 
      WriteString[ostrm, "\\State[", stlab[[p]], "]{", statecoord[p],
        "}{", vlab[[p]], "}\n"];
    
    writeTrans[{p_, p_}] := 
      With[{labs = trf[{p, p}]},
        WriteString[ostrm, "\\LoopN{", vlab[[p]], "}"];
        writeStringList[ostrm, alph[[labs]] ];
        ];

    writeTrans[{p_, q_}] := 
      With[{labs = trf[{p, q}]},
        WriteString[ostrm, "\\EdgeL{", vlab[[p]], "}{", vlab[[q]], "}"];
        writeStringList[ostrm, alph[[labs]] ];
        ];
    
    writeStringList[os_, L_List, sep_String:","] :=
      (
        WriteString[os, "{"];
        WriteString[os, #, sep] & /@ Most[L];
        WriteString[os, Last[L]];
        WriteString[os, "}\n"]
        );
    
    	n = Size[M];
    	{k, alph} = Alphabet[M, Full -> True];
    	vlab = FromCharacterCode /@ Range[65, 64 + n];
    	stlab = StringReplace[ToString[#], " " -> ""] & /@ States[M] ;
    	tr = Transitions[ToFA[M]];
    	trf[{_, _}] := {};
    	Scan[  AppendTo[trf[{#[[1]], #[[3]]}], #[[2]]] &, tr ];
    	st = Union[Delete[#, {2}] & /@ tr];
    	ostrm = OpenWrite[fl <> ".vcs"];
    	writeStates /@ Range[n];
    	WriteString[ostrm, "%\n"];
    	writeTrans /@ st;
    	Close[ostrm]
    ];


(* ::Section::Closed:: *)
(*Regex*)


(* ::Subsection::Closed:: *)
(*Regular Expressions*)


prot=Unprotect[Empty,NonCommutativeMultiply];
ClearAll[Empty,RET,REP,RES]
SetAttributes[NonCommutativeMultiply,Listable];
Protect[Evaluate[prot]];

(*$defaultregexalphabet = Alpha[75,Printable->FromCharacterCode/@ Range[48,122]];*)
$defaultregexalphabet = Alpha[26,Printable->Alphabet[]];

RegularExpressionPrettyPrint[ exp_ ] := 
	 ReplaceRepeated[ exp, 
       {""->\[Epsilon],Eps->\[Epsilon],REP->Plus,RET->NonCommutativeMultiply,RES[x_]:> Power[x,"*"]}, 
     MaxIterations->100 ] ;

ToRegularExpression[exp_] := exp /. {NonCommutativeMultiply-> RET,Plus-> REP,Times[_Integer,x_]-> x };


(* ::Subsection::Closed:: *)
(*BEFA*)


Clear[makeBEFA,BEFA,BEFAToFA]

shiftrans[ n_][{x_,s_,z_} ] := {x + n, s, z + n};
befasize[M_BEFA] := M[[1]];
befatrans[M_BEFA] := M[[2]];

makeBEFA[Empty]      := BEFA[2,{}];
makeBEFA[_?EpsQ]  := BEFA[2,{{1,0,2}}];
makeBEFA[xx_String] := Module[ {L,n},
     L = ToIndex[xx,$regexalphabet];
     n = Length[L] + 1;
     BEFA[ n, Thread[ {Range[n-1], L, Range[2,n]} ] ]
];
makeBEFA[xx_REP]    := unionBEFA @@ (makeBEFA /@ xx);
makeBEFA[xx_RET]    := concatenateBEFA @@ (makeBEFA /@ xx);
makeBEFA[xx_RES]    := starBEFA[ makeBEFA[First[xx]] ];


unionBEFA[ M1_BEFA, M2_BEFA ] := 
Module[ {n1, n2, T1, T2},
     n1 = befasize[ M1 ];
     n2 = befasize[ M2 ];
     T1 = befatrans[M1] /. {x_Integer,s_,n1} :> {x,s,n1+n2-2} ;
     T2 = Map[ shiftrans[n1-2], befatrans[M2] ]; 
     T2 = T2 /. {n1-1,s_,y_Integer} :> {1,s,y};
     BEFA[ n1+n2-2, Union[T1,T2] ]
];
unionBEFA[ M1_BEFA, M2_BEFA, z___BEFA ] := unionBEFA[ unionBEFA[ M1, M2 ], z ];

concatenateBEFA[ M1_BEFA, M2_BEFA ] := 
Module[ {n1,n2,T1,T2},
     n1 = befasize[M1];
     n2 = befasize[M2];
     T1 = befatrans[M1];
     T2 = Map[ shiftrans[n1-1], befatrans[M2] ];
     BEFA[ n1+n2-1, Union[T1,T2] ]
];
concatenateBEFA[ M1_BEFA, M2_BEFA, z___BEFA ] := concatenateBEFA[ concatenateBEFA[ M1, M2 ], z ];

starBEFA[ M_BEFA ] := 
Module[ {n, Tr},
     n = befasize[ M ];
     Tr = Map[ shiftrans[1], befatrans[M] ];
     BEFA[ n+2, Union[Tr,{{1,0,2},{n+1,0,2},{2,0,n+2}}] ]
];

BEFAToFA[ M_BEFA, K_Alpha:$defaultregexalphabet ] :=
    FA[  TSys[ befasize[M], AlphabetCount[K], Length[ befatrans[M]],  befatrans[M], K], {1}, {befasize[M]} ] ;



(* ::Subsection::Closed:: *)
(*NNFA*)


(* ::Text:: *)
(*McNaughton-Yamada style normal NFAs*)


Clear[makeNNFA,NNFA,NNFAToFA]

NNFAToFA[ M_NNFA, K_Alpha:$defaultregexalphabet ] := 
Module[ {insert,tr},
     insert[{p_,q_}] := {p,M[[3,q]],q};
     tr = insert /@ Join[ CartesianProduct[{0},M[[4]] ], Zwei[M] ];
     tr = MapTransitions[  tr, #+1&, ToIndex[#,K][[1]]& ];
     FA[  TSys[ First[M] + 1,AlphabetCount[K],Length[tr],tr, K ],
         {1},
         If[ M[[6]]==0, M[[5]], Union[ M[[5]], {0} ] ] + 1
     ]
];

makeNNFA[_?EpsQ] := NNFA[0,{},{},{},{},1];
makeNNFA[Empty]     := NNFA[0,{},{},{},{},0];
makeNNFA[ xx_String ]     := 
With[ {n=StringLength[xx]},
      NNFA[ n, Partition[Range[n],2,1], Characters[xx], {1}, {n}, 0 ]
];
makeNNFA[xx_REP]    := unionNNFA @@ (makeNNFA /@ xx);
makeNNFA[xx_RET]    := concatenateNNFA @@ (makeNNFA /@ xx);
makeNNFA[xx_RES]    := starNNFA[ makeNNFA[First[xx]] ];
(*sizeNNFA[ e_?WordQ ] :=  StringLength[e];
sizeNNFA[ e_REP|e_RET ] := Total[sizeNNFA/@e];
sizeNNFA[ e_RES ]    := sizeNNFA[e[[1]]];
*)

unionNNFA[ M_NNFA ] := M; 
unionNNFA[ M1_NNFA, M2_NNFA ] := 
Module[ {n=M1[[1]], I1=M1[[4]], F1=M1[[5]], I2, F2},
     I2 = M2[[4]] + n;
     F2 = M2[[5]] + n;
     NNFA[    
            n + M2[[1]],
            Union[ M1[[2]], M2[[2]] + n ],
            Join[ M1[[3]], M2[[3]] ],
            Union[ I1, I2 ],
            Union[ F1, F2 ],
            Max[ M1[[6]], M2[[6]] ] 
     ]
];
unionNNFA[ M1_NNFA, M2_NNFA, M___NNFA ] :=  
unionNNFA[ unionNNFA[ M1, M2 ], M ];

concatenateNNFA[ M_NNFA ] := M; 
concatenateNNFA[ M1_NNFA, M2_NNFA ] := 
Module[ {n=M1[[1]], I1=M1[[4]], F1=M1[[5]], I2, F2},
     I2 = M2[[4]] + n;
     F2 = M2[[5]] + n;
     NNFA[    
	   n + M2[[1]],
	   Union[ M1[[2]], M2[[2]]+n, CartesianProduct[F1,I2] ],
	   Join[ M1[[3]], M2[[3]] ],
	   If[M1[[6]]==0, I1, Union[ I1, I2 ] ],
	   If[M2[[6]]==0, F2, Union[ F1, F2 ] ],
	   M1[[6]] M2[[6]]
     ]
];
concatenateNNFA[ M1_NNFA, M2_NNFA, M___NNFA ] := 
concatenateNNFA[ concatenateNNFA[ M1, M2 ], M ];
starNNFA[ M_NNFA ] := 
    ReplacePart[ M, {2->Union[ Zwei[M], CartesianProduct[M[[5]],M[[4]]] ], 6->1} ]; 



(* ::Subsection::Closed:: *)
(*Conversion to FA*)


ClearAll[RegularExpressionToFA]
Options[RegularExpressionToFA]={Thompson->False,EpsElim->True,DoTrace->False};
RegularExpressionToFA[  regraw_, K_Alpha:$defaultregexalphabet, opts:OptionsPattern[] ] := 
Module[ {M,reg},
	reg = ToRegularExpression[regraw];
	$regexalphabet = K;
	M = If[ OptionValue[Thompson],
		makeBEFA[reg],
		makeNNFA[reg]
	];
	If[ OptionValue[DoTrace], Return[M] ];
	M = If[  Head[M]===BEFA, 
		BEFAToFA[ M, K ],
		NNFAToFA[ M, K ]
	];	
	If[ OptionValue[Thompson]&& OptionValue[EpsElim],
		EpsilonEliminationFA[M],
		M
	]
];
	


(* ::Subsection::Closed:: *)
(*Conversion to Regex*)


ClearAll[ToRegularExpressionFA]
Options[ToRegularExpressionFA]={RegexSimp->True,DoTrace->False};
ToRegularExpressionFA[MM_?FSMQ,opts:OptionsPattern[]] := 
Module[{M,k,n,A,tran,simp,L,trace,trc},
    simp=OptionValue[RegexSimp];
    trace=OptionValue[DoTrace];
    trc={};
    M=ToFA[MM];
    n=StateCount[M];
    A=AlphabetList[GetAlphabet[M]];
    tran=MapTransitions[TransitionList[M],Identity,A[[#1]]&];

    L[0,p_,q_] := L[0,p,q] =
    Module[{lab,exp},
        lab=Zwei/@Cases[tran,{p,_,q}];
        exp = If[lab==={},
                 If[p==q,"",Empty],
                 If[p==q,
                        REP[Eps,Sequence@@lab],
                        If[Length[lab]>1,REP@@lab,lab[[1]]]
                 ]
                ];
        If[trace,AppendTo[trc,{{0,p,q},exp}]];
        If[simp,RegularExpressionSimplify[exp],exp]
    ];


    L[s_,p_,q_] := L[s,p,q] =
    Module[{exp},
        exp = REP[L[s-1,p,q],RET[L[s-1,p,s],RES[L[s-1,s,s]],L[s-1,s,q]]];
        If[ simp, exp = exp//. $regexrules];
        If[ trace, AppendTo[trc,{{s,p,q},exp}]];
        If[ simp, RegularExpressionSimplify[exp],exp]
    ];

    exp = REP@@( L[n,#1[[1]],#1[[2]]]&) /@ CartesianProduct[InitialStateList[M],FinalStateList[M]];
    If[simp,exp=RegularExpressionSimplify[exp]];
    If[trace,{exp,trc},exp] /. ""->Eps
];


(* ::Subsection::Closed:: *)
(*Simplification*)


(* ::Text:: *)
(*Must turn off $Post before use.*)
(*Operates on "", not Eps.*)


ClearAll[RegularExpressionSimplify]
RegularExpressionSimplify[exp_]:= 
With[{sexp=exp/. {\[Epsilon]->"","Eps"->"",Eps->"",\[EmptySet]->Empty}}, 
    ReplaceRepeated[sexp,$regexrules,MaxIterations->100]] /. ""->Eps;


$regexrules = Dispatch[{
     REP[x___,REP[z__],y___] :>  REP[x,z,y], 
     REP[x___,Empty,y___] :>  REP[x,y],
     REP[x___,u_,y___,u_,z___] :>  REP[x,u,z,y],
     RET[x___,RET[z___],y___] :>  RET[x,z,y], 
     RET[___,Empty,___] ->  Empty,
     RET[x___,"",y___] :> RET[x,y],
     REP[] ->  Empty,
     RET[] ->  "",
     REP[xx_] :>   xx,
     RET[xx_] :>  xx,
     RES[Empty] -> "",
     RES[""] -> "",
     RES[RES[x_]] :> RES[x],
     REP[x___,u_,z___,RES[u_],y___] :>  REP[x,z,RES[u],y],
     REP[x___,RES[u_],z___,u_,y___] :>  REP[x,RES[u],z,y],
     RES[REP[x___,"",y___]] :>  RES[REP[x,y]],
     RET[x___,REP["",y__],RES[y__],z___] :>  RET[x,RES[y],z] ,
     RET[x___,RES[y__],REP["",y__],z___] :>  RET[x,RES[y],z], 
     REP[x___,RES[y__],z___,"",u___] :>  REP[x,RES[y],z,u],
     REP[x___,"",z___,RES[y__],u___] :>  REP[x,RES[y],z,u],
     RET[x___,a_String,b_String,y___] :> RET[x,StringJoin[a,b],y]
}];


(* ::Section::Closed:: *)
(*Words*)


(* ::Text:: *)
(*Revive for v11?*)


(*ClearAll[Concatenate]
Concatenate[Eps|Global`\[Epsilon]] := "";
Concatenate[x_String]  := x;
Concatenate[x:{___Integer}]  := x;
Concatenate[Eps|Global`\[Epsilon],y_] := y;
Concatenate[x_,Eps|Global`\[Epsilon]] := x;
Concatenate[xx___,Eps|Global`\[Epsilon],yy___] := Concatenate[xx,yy];
Concatenate[ {xx___,Eps|Global`\[Epsilon],yy___} ] := Concatenate[xx,yy];
Concatenate[xx__String] := StringJoin[xx];
Concatenate[xx:{__String}]  := StringJoin@@xx;
Concatenate[xx__List,yy_List] := Apply[ Concatenate, CartesianProduct[xx,yy], 1 ];
Concatenate[xx_List,yy_List, zz___List ] := Concatenate[  Concatenate[xx,yy],zz];*)


ClearAll[ParikhVector]
ParikhVector[A_Alpha][s_String] := 
With[ {sc= Characters[s]}, 
	AssociationMap[ Count[sc,#]&, AlphabetList[A] ] ];

ParikhVector[A_Alpha][L:{___String}] := 
With[ {al = AlphabetList[A]}, 
	Table[ AssociationMap[ Count[Characters[s],#]&, al ], {s,L} ] ];


(* ::Section::Closed:: *)
(*Divisibility*)


ClearAll[DivisibilityDFA]
DivisibilityDFA::noimp = "Option `1` currently not implemented.";

Options[DivisibilityDFA] = {Direction -> Forward, Full -> True, SizeOnly -> False};

DivisibilityDFA[m_Integer?Positive, 1, opts : OptionsPattern[]] :=
  If[  OptionValue[SizeOnly],
   	Message[DivisibilityDFA::noimp, "SizeOnly"]; Return[$Failed];
   	sizeDivisibility[m, 1, opts],
   	DFA[ TSys[ m, 1, m, {RotateLeft[Range[m]]}, Alpha[1], Type -> Matrix, States -> (Range[m] - 1)], 1, {1} ]
   ];
DivisibilityDFA[m_Integer?Positive, B_Integer?Positive, opts : OptionsPattern[]] :=
  (
   If[OptionValue[SizeOnly],
    	Message[DivisibilityDFA::noimp, "SizeOnly"]; Return[$Failed];
    	Return[sizeDivisibility[m, B, opts]]
    ];
   If[OptionValue[Full],
    	Return[canonicalDivisibilityDFA[m, B, OptionValue[Direction]]]];
   Message[DivisibilityDFA::noimp, "Full"]; Return[$Failed];
   compressedDivisibilityDFA[m, B, OptionValue[Direction]]
   );
canonicalDivisibilityDFA[m_Integer?Positive, B_Integer?Positive, dir_] :=
  Module[{Q, del},
   If[Forward === dir,
    Return[ DFA[ TSys[ m, B, m B, Thread[Partition[Mod[Range[m B] - 1, m], B]] + 1, 
       	Alpha[B, Type -> Digit], Type -> Matrix, States -> Range[0, m - 1]], 1, {1}]]
    ];
   (* only for full *)
   (*Q=CartesianProduct[Range[0,m-1],Closure[1,{Mod[B #1,m]&}]];*)
   del[a_][{p_, q_}] :=  Mod[ { p + (a - 1) q, q B }, m ];    (* 1-indexing *)
   GenerateFA[ {0, 1},
    Table[  Evaluate[del[a]], {a, B} ], Alpha[B, Type -> Digit],
    First[#] == 0 &, Type -> DFA, Normalize -> Automatic ]
   ];


(* ::Section::Closed:: *)
(*Footer*)


Null
