(* ::Package:: *)

(* ::Title:: *)
(*Plot*)


(* ::Text:: *)
(*Author:		Klaus Sutner*)
(*Affiliation: 	Carnegie Mellon University*)
(*email:       		sutner@cs.cmu.edu*)
(*url:         		http://www.cs.cmu.edu/~sutner*)
(*version:     	11.0*)


(* ::Section::Closed:: *)
(*Header*)


BeginPackage["Automata11`plot`"];

Begin["`Private`"];
End[];

EndPackage[];


(* ::Section::Closed:: *)
(*Graph Layouts, Themes*)


$GraphEmbeddings = {"SpringEmbedding", "SpringElectricalEmbedding", "SpectralEmbedding", 
"PlanarEmbedding", "HighDimensionalEmbedding", "TutteEmbedding", "BipartiteEmbedding", 
"CircularEmbedding", "CircularMultipartiteEmbedding", "DiscreteSpiralEmbedding", 
"GridEmbedding", "LinearEmbedding", "MultipartiteEmbedding", "SpiralEmbedding", 
"StarEmbedding", "BalloonEmbedding", "RadialEmbedding", "LayeredDigraphEmbedding", 
"LayeredEmbedding"};

TryLayouts[ gr_Graph ] := 
  	Partition[SetProperty[gr, {GraphLayout -> #, PlotLabel -> #}] & /@ 
  		$GraphEmbeddings, UpTo[4] ] // GraphicsGrid;

$PlotThemes = {"Business", "Detailed", "Marketing", "Scientific", "Web", "Classic", 
"Minimal", "Monochrome"};

TryThemes[ G_Graph ] := 
 	Partition[ Table[ SetProperty[G, {PlotTheme -> tm, PlotLabel -> tm}], {tm, $PlotThemes} ], 4 ] // GraphicsGrid


(* ::Section::Closed:: *)
(*PlotGrid*)


Unprotect[PlotGrid];
ClearAll[PlotGrid]
SyntaxInformation[PlotGrid] = {"ArgumentsPattern"->{__, OptionsPattern[]}};
Options[PlotGrid] =
{PlotStyle->Directive[GrayLevel[0.8],Thickness[Medium]],
GridOffset->{0,0},Full->All};

PlotGrid[{m_Integer,n_Integer},opts:OptionsPattern[] ] :=
    PlotGrid[m,n,opts];
PlotGrid[m_Integer,n_Integer,opts:OptionsPattern[] ] :=
    Module[ {ps,xoff,yoff,full,lines},
        ps = OptionValue[PlotStyle];
        {xoff,yoff} = OptionValue[GridOffset];
        full = OptionValue[Full];
        lines = Switch[full,
        All,Join[Line/@Table[{{i-xoff,1-yoff},{i-xoff,m-yoff}},{i,n}],
                        Line/@Table[ {{1-xoff,i-yoff},{n-xoff,i-yoff}},{i,m}]],
        _List,{Map[Line,full,{-3}]}
        ];
        Graphics[Join[{ps},lines],SelectOptions[Graphics,opts]]
    ];


(* ::Section::Closed:: *)
(*PlotMatrix*)


$RasterColors = {RGBColor[0.8,1.,1.],RGBColor[0,0.2,1.],RGBColor[1.,1.,0],
RGBColor[0.2,1.,0.2],RGBColor[1.,0.,0.],RGBColor[0.,1.,1.],
RGBColor[0.53,0.26,0.12],RGBColor[1.,0,1.],RGBColor[0.75,0.75,0.75],RGBColor[0.,0.,0.]};
$RasterColorRules = Thread[ Range[0,Length[$RasterColors]-1] -> $RasterColors ];

Unprotect[PlotMatrix];
ClearAll[PlotMatrix]
SyntaxInformation[PlotMatrix] = {"ArgumentsPattern"->{_, OptionsPattern[]}};

Options[PlotMatrix] = Join[{Frame->True,
      MultiColumns->False,NRows->80,NoRest->True,
	  ColorFunction->None,ColorRules:>$RasterColorRules}, Options[ArrayPlot]];
SetOptions[PlotMatrix,MeshStyle->{{GrayLevel[0.8],Thin},{GrayLevel[0.8],Thin}}];

PlotMatrix[mm_?MatrixQ, opts:OptionsPattern[]] :=
Module[ {colf,gr},
	If[ OptionValue[MultiColumns],
		Return[PlotMatrixBlocks[mm,opts] ]
	];

	If[ (colf = OptionValue[ColorFunction])=!=None,
		ArrayPlot[mm,ColorFunction->colf,Frame->OptionValue[Frame],
					SelectOptions[ArrayPlot,opts]]
			(*else*),
		ArrayPlot[mm,ColorRules->OptionValue[ColorRules],Frame->OptionValue[Frame],
					SelectOptions[ArrayPlot,opts]]
	]
];


ClearAll[PlotMatrixBlocks]
Options[PlotMatrixBlocks] = Options[PlotMatrix];
PlotMatrixBlocks[L_?MatrixQ,opts:OptionsPattern[] ] :=
    Module[ {part,nr },
        nr = OptionValue[NRows];
        If[ OptionValue[NoRest],
            part = Partition[L,nr],
            part = Partition[L,nr,nr,1,{}]
        ];
       GraphicsRow[PlotMatrix[#,SelectOptions[PlotMatrix, opts, {MultiColumns}]]&/@part,Alignment->Top ]
    ];
      


(* ::Section::Closed:: *)
(*PlotFA*)


$defaultedgecolor = {Red,Green,Blue,Orange,Cyan,Black};
ClearAll[ColorVertices,LabelEdges,PlotFA]

Options[PlotFA] = {Trim->True,Colors->True,VertexSize->Automatic,VertexLabels->None,PlotTheme->"Default",
ImagePadding->All};
PlotFA[ MM_?FSMQ, opts:OptionsPattern[] ] := 
Module[ {g,M},
	M = If[ OptionValue[Trim], TrimFA[MM], MM ];
	g = TransitionsToGraph[M];
	g = ColorVertices[ g, InitialStateList[M], FinalStateList[M],SelectOptions[ColorVertices,opts] ];
	g = LabelEdges[ g, TransitionList[M], GetAlphabet[MM], SelectOptions[LabelEdges,opts] ];
	SetProperty[ g,  {VertexLabels-> OptionValue[VertexLabels], GraphLayout->Automatic, 
					PlotTheme->OptionValue[PlotTheme],ImagePadding->OptionValue[ImagePadding]}]
];

Options[ColorVertices] = {VertexSize->Automatic};
ColorVertices[ g_Graph, ini_List, fin_List, opts:OptionsPattern[] ] := 
Module[ {Vi,Vf,Vif,gr,vs},
    vs = OptionValue[VertexSize];
	Vi = Complement[ini,fin];
	Vf = Complement[fin,ini];
	Vif = Intersection[ini,fin];
	Vother = Complement[ VertexList[g], ini, fin ];
	gr = Fold[  SetProperty[ {#1,#2}, {VertexSize->vs,VertexStyle->LightYellow}]&, g, Vi ];
	gr = Fold[  SetProperty[ {#1,#2}, {VertexSize->vs,VertexStyle->LightGreen}]&, gr, Vif ];
	gr = Fold[  SetProperty[ {#1,#2}, {VertexSize->vs,VertexStyle->LightBlue}]&, gr, Vf ];
	gr = Fold[  SetProperty[ {#1,#2}, {VertexSize->vs,VertexStyle->LightGray}]&, gr, Vother ];
	gr
];

Options[LabelEdges] = {Colors->None};
LabelEdges[ g_Graph, tr_List, K_Alpha, opts:OptionsPattern[] ] := 
Module[ {ed,gr,col,p,q,k},
	gr = g;
    k = AlphabetCount[K];

(* colored edges *)
	If[ (col = OptionValue[Colors])=!=None, 
        If[ col === Automatic, col = $defaultedgecolor ];
        PrependTo[ col, Dashed ];
		Do[ 
			ed = Cases[tr,{p_,s,q_} :> DirectedEdge[p,q] ];
			gr = Fold[  SetProperty[ {#1,#2}, {EdgeStyle->col[[s+1]]}]&, gr, ed ]
			,
			{s,0,k}
		];
		Return[gr]
	];

(* labeled edges *)

	ed = GatherBy[tr, {TransitionSource[#1], TransitionTarget[#1]}& ]; 
	ed = auxCombineLabels[#,K]& /@ ed; 
	gr = Fold[  auxGenerateLabels, gr, ed ]
];

auxCombineLabels[ L:{_List...}, K_Alpha ] := 
    With[ {p = L[[1,1]], q = L[[1,-1]], ll = Sort[Zwei/@L]},
        If[ MemberQ[ll,0], 
              {p,StringJoin[ToString[\[Epsilon]],ToWord[ ll, K ]],q},
              {p,ToWord[ ll, K ],q}
        ]
    ];

auxGenerateLabels[ g_Graph, {p_,"",q_} ] :=  SetProperty[ {g,DirectedEdge[p,q]}, 
		{EdgeStyle->Dashed, EdgeLabels->\[Epsilon],EdgeLabelStyle->Directive[15,Bold,Background->White]}];
auxGenerateLabels[ g_Graph, {p_,s_String,q_} ] :=  SetProperty[ {g,DirectedEdge[p,q]}, 
		{EdgeLabels->s,EdgeLabelStyle->Directive[15,Bold,Background->White]}];


(* ::Section::Closed:: *)
(*Miscellany*)


(*Clear[binTrees]
binTrees[1] = {{1}};
binTrees[n_] :=
    binTrees[n] =
    FlattenOne@Table[ combineTrees /@ CartesianProduct[binTrees[i], binTrees[n-i]], {i,n-1}];

Clear[combineTrees]
combineTrees[ {t1_, t2_} ] :=
    Module[ {k},
        k = Max[ Level[t1,{-1}]];
        DeleteCases[Join[  {1->2,1->k+2},t1 /. {x_Integer :>  x+1},t2/.{x_Integer :>  x+k+1} ],_Integer]
    ];*)


prefixGraph[L_List] :=
    Union@@(Apply[ Rule,Partition[WordPrefixes[ # ],2,1], {1} ]& /@ L);
plotFrontier[L_List] :=
    Module[ {W,prf},
        W = prefixGraph[ Words[Max[StringLength/@ L]]];
        prf = Union@@WordPrefixes[L];
        TreePlot[W,VertexRenderingFunction->
        (Flatten[{If[ MemberQ[prf,#2],
                      If[ MemberQ[ L,#2],
                          {PointSize[0.02],LightRed},
                          {PointSize[0.015],Black}
                      ],
                      {PointSize[0.015],GrayLevel[0.9]}
                  ],Point[#1]}]&),
        LayerSizeFunction->(0.3&)]
    ];

prefixGraph[L_List] :=
    Union@@(Apply[ Rule,Partition[WordPrefixes[ # ],2,1], {1} ]& /@ L);
plotFrontier[L_List] :=
    Module[ {W,prf},
        W = prefixGraph[ Words[Max[StringLength/@ L]]];
        prf = Union@@WordPrefixes[L];
        TreePlot[W,VertexRenderingFunction->
        (Flatten[{If[ MemberQ[prf,#2],
                      If[ MemberQ[ L,#2],
                          {EdgeForm[{Black,Thick}],LightRed,Disk[#1,0.1],Black,Text[ToString[ delta[1,#2 ]],#1]},
                          {EdgeForm[Black],LightYellow,Disk[#1,0.1],Black,Text[ToString[ delta[1,#2 ]],#1]}
                      ],
                      {PointSize[0.015],GrayLevel[0.9],Point[#1]}
                  ]}]&),
        AspectRatio->1/2]
    ];


ClearAll[PlotFrontierDFA]

PlotFrontierDFA[  \[CapitalGamma]_List ] :=
    Module[ {int,AA,A,B,coords},
        int = prefixGraph[First /@ \[CapitalGamma]];
        AA = Select[ int, MemberQ[ First/@ \[CapitalGamma], #[[2]] ]& ];
        B = Complement[int,AA];
        A = AA/. \[CapitalGamma];
        coords =  ExtractCoordinates[
        TreePlot[  B, Top,DirectedEdges->True, VertexLabeling->True, LayerSizeFunction->(0.4&)  ]];
        LayeredGraphPlot[  DeleteDuplicates@Join[B,A], Top,
        VertexLabeling->False,
        VertexCoordinateRules->coords,
        EdgeRenderingFunction-> edgeGray[{0.01,0.05},A/. Rule->List] ]
    ];





(* ::Section::Closed:: *)
(*Footer*)


Null
