(* ::Package:: *)

(* ::Title:: *)
(*Semigroups*)


(* ::Text:: *)
(*Author:		Klaus Sutner*)
(*Affiliation: 	Carnegie Mellon University*)
(*email:       		sutner@cs.cmu.edu*)
(*url:         		http://www.cs.cmu.edu/~sutner*)
(*version:     	11.0*)


(* ::Section::Closed:: *)
(*Header*)


BeginPackage["Automata11`sgroup`"];

Begin["`Private`"];
End[];

EndPackage[];


(* ::Section::Closed:: *)
(*SGT*)


(* ::Text:: *)
(*Transformations  [n] \[LongRightArrow] [n]*)


ClearAll[KernelSGT, ImageSGT, RankSGT, FromKernelImageSGT, IdempotentQSGT,SelectorQ,OneSGT]

KernelSGT[ a_SGT]  := Flatten[Position[First[a], #]] & /@ ImageSGT[a];
ImageSGT[ a_SGT]  := Union[First[a]];
RankSGT[ a_SGT]  := Length[ImageSGT[a]];
FromKernelImageSGT[ker_List, im_List] := SGT @@ (Last /@ Sort@FlattenOne[Thread /@ Pairs[ker, im]]);
IdempotentQSGT[a_SGT] :=  SemigroupProduct[a, a] === a;
IdempotentQSGT[L:{__SGT}] :=  Select[L,SemigroupProduct[#,#] === #& ];
SelectorQ[im_List,ker_List]:=Union[Length/@(Intersection[#,im]&/@ker)]==={1};
OneSGT[gen:{__SGT}] := OneSGT[ SemigroupDegree[gen]];
OneSGT[d_Integer?Positive] := SGT[Range[d]];ClearAll[KernelSGT, ImageSGT, RankSGT, FromKernelImageSGT, IdempotentQSGT,SelectorQ,OneSGT]
KernelSGT[ a_SGT]  := Flatten[Position[First[a], #]] & /@ ImageSGT[a];
ImageSGT[ a_SGT]  := Union[First[a]];
RankSGT[ a_SGT]  := Length[ImageSGT[a]];
FromKernelImageSGT[ker_List, im_List] := SGT @@ (Last /@ Sort@FlattenOne[Thread /@ Pairs[ker, im]]);
IdempotentQSGT[a_SGT] :=  SemigroupProduct[a, a] === a;
IdempotentQSGT[L:{__SGT}] :=  Select[L,SemigroupProduct[#,#] === #& ];
SelectorQ[im_List,ker_List]:=Union[Length/@(Intersection[#,im]&/@ker)]==={1};
OneSGT[gen:{__SGT}] := OneSGT[ SemigroupDegree[gen]];
OneSGT[d_Integer?Positive] := SGT[Range[d]];



(* ::Section::Closed:: *)
(*Green's Relations*)


ClearAll[GreenLeftQ,GreenRightQ,GreenDQ,GreenHQ,DClassSGT]

GreenLeftQ[gen:{__SGT}][ x_SGT, y_SGT ] := 
With[ {S = SemigroupElements[gen]},
With[ {xx = PositionOne[S,x],yy = PositionOne[S,y]},
	PositionOne[ SemigroupGreenLeft[gen], xx ] === PositionOne[ SemigroupGreenLeft[gen], yy ] ] ];
GreenRightQ[gen:{__SGT}][ x_SGT, y_SGT ] := 
With[ {S = SemigroupElements[gen]},
With[ {xx = PositionOne[S,x],yy = PositionOne[S,y]}, 
	PositionOne[ SemigroupGreenRight[gen], xx ] === PositionOne[ SemigroupGreenRight[gen], yy ] ] ];
GreenDQ[gen:{__SGT}][ x_SGT, y_SGT ] := 
With[ {S = SemigroupElements[gen]},
With[ {xx = PositionOne[S,x],yy = PositionOne[S,y]}, 
	PositionOne[ SemigroupDClasses[gen], xx ] === PositionOne[ SemigroupDClasses[gen], yy ] ] ];GreenHQ[gen:{__SGT}][ x_SGT, y_SGT ] := GreenRightQ[gen][x,y] && GreenLeftQ[gen][x,y];

DClassSGT[gen:{__SGT}][a_SGT] := 
Module[ {S1,rk,Ra,La,H,LR},
	S1 = SemigroupToMonoid[gen];
	rk = RankSGT[a];
	Ra  = DeleteDuplicates[ Pairs[a**S1,S1], First ];
	Ra = Select[ Ra, RankSGT[First[#]]==rk& ];
	H    = Union@Select[First /@ Ra,(ImageSGT[#]==ImageSGT[a])& ]; 
	Ra  = Last /@ DeleteDuplicatesBy[Pairs[ImageSGT[First[#]]& /@ Ra, Last /@ Ra ],First];
	La  = DeleteDuplicatesBy[Pairs[S1**a,S1],First ];
	La  = Select[La,(RankSGT@First[#]==rk)& ];
	La  = Last/@DeleteDuplicatesBy[Pairs[Sort@KernelSGT@First[#]& /@ La, Last /@ La],First];
	LR = CartesianProduct[La,Ra];
	LR = Select[ LR, SelectorQ[ImageSGT[a**#[[2]]],KernelSGT[#[[1]]**a]]&];
	{ Union[First  /@ LR],H,Union[Last /@ LR] }
];

Clear[DClassToList]
Options[DClassToList] = {Full->True};
DClassToList[ {L_List,H_List,R_List},opts:OptionsPattern[] ] := 
	With[ {tab = Outer[#1**H**#2&,L,R]},
		If[ OptionValue[Full], tab, Sort@Flatten[tab] ] ];


(* ::Section::Closed:: *)
(*Semigroup*)


ClearAll[SemigroupElements,SemigroupRegularElements,SemigroupDClasses,SemigroupGreenLeft,SemigroupGreenRight,SemigroupWitnesses,SemigroupEquations,SemigroupOrder,SemigroupDegree];

SemigroupElements[M_DFA] :=SemigroupElements[ToTransformationsDFA[M]];SemigroupDClasses[M_DFA] :=SemigroupDClasses[ToTransformationsDFA[M]];SemigroupGreenLeft[M_DFA] :=SemigroupGreenLeft[ToTransformationsDFA[M]];SemigroupGreenRight[M_DFA] :=SemigroupGreenRight[ToTransformationsDFA[M]];SemigroupWitnesses[M_DFA] :=SemigroupWitnesses[ToTransformationsDFA[M]];SemigroupEquations[M_DFA] :=SemigroupEquations[ToTransformationsDFA[M]];
SemigroupOrder[M_DFA] :=SemigroupOrder[ToTransformationsDFA[M]];

SemigroupClear[M_DFA] :=SemigroupClear[ToTransformationsDFA[M]];
SemigroupClear[gen:{__SGT}] := 
(
	Quiet@Unset[ {SemigroupElements[gen],
SemigroupDClasses[gen],
SemigroupGreenLeft[gen],
SemigroupGreenRight[gen],
SemigroupWitnesses[gen],
SemigroupEquations[gen]}];
	Null
);
SemigroupRegularElements[gen:{__SGT}] := 
With[ {S = SemigroupElements[gen]}, 
	Select[ S, MemberQ[#**S**#,#]&] ];
	
ClearAll[SemigroupToMonoid]
SemigroupToMonoid[gg:{SGT__}] := 
Module[ {S = SemigroupElements[gg]},
	If[ Select[ Select[S, IdempotentQSGT ], (#**gg===gg** #===gg)& ] =!= {},
	S,Prepend[ S, OneSGT[gen] ]]
];  

ClearAll[TransformationSemigroup]
TransformationSemigroup[M_DFA] := TransformationSemigroup[
ToTransformationsDFA[M]];
TransformationSemigroup[gen:{__SGT}]  :=
(
	SemigroupClear[gen];
 {SemigroupElements[gen],
{SemigroupGreenLeft[gen],
SemigroupGreenRight[gen],
SemigroupDClasses[gen]},
SemigroupWitnesses[gen],
SemigroupEquations[gen]} = GenerateSemigroup[ gen, GreenRel->True ];
SemigroupOrder[gen] = Length[SemigroupElements[gen]];

Echo[ SemigroupOrder[gen], "semigroup order: " ];
Null
)
Clear[SemigroupMultiplicationTable]
Options[SemigroupMultiplicationTable] = {Full->False};
SemigroupMultiplicationTable[S_,opts:OptionsPattern[] ] :=
With[ {tab = Table[ SemigroupProduct[s,t], {s,S}, {t,S} ]},
	If[ OptionValue[Full], 
		tab,
		Map[ PositionOne[S,#]&, tab, {2} ]
	]
];
Clear[PrintRewrite]
PrintRewrite[wit_List,A_Alpha][{p_,a_Integer,q_}] :=  Quiet@StringJoin[wit[[p]] ,"\[CenterDot]" ,AlphabetList[A][[a]]," \[Rule] " , wit[[q]] ] ; 
PrintRewrite[wit_List,A_Alpha][L:{{_,_Integer,_}...}] :=  PrintRewrite[wit,A] /@ L;

Options[OneElementSG] = {Direction->All};
OneElementSG[gen:{__SGT},opts:OptionsPattern[]] :=
With[{idem = Select[SemigroupElements[gen], IdempotentQSGT ]},
	Switch[ OptionValue[Direction],
	Left, Select[ idem, (#**gen===gen)& ],
	Right, Select[ idem, (gen** #===gen)& ],
	All,  Select[ idem, (#**gen===gen** #===gen)& ]
	]
];

Options[NullElementSG] = {Direction->All};
NullElementSG[gen:{__SGT},opts:OptionsPattern[]] :=
With[{idem = Select[SemigroupElements[gen],IdempotentQSGT ],d= SemigroupDegree[gen]},
	Switch[ OptionValue[Direction],
	Left, Select[ idem, (#**gen===Table[#,{d}])& ],
	Right, Select[ idem, (gen** #===Table[#,{d}])& ],
	All,  Select[ idem, (#**gen===gen** #===Table[#,{d}])& ]
	]
];


(* ::Text:: *)
(*Only 6 colors, add colorfunction.*)
(*Does not handle multi-edges properly (Mathematica bug).*)


Options[SemigroupCayleyGraph] = {Full->False,Cayley->False,GraphLayout-> Inherited};
SemigroupCayleyGraph[gen:{__SGT},opts:OptionsPattern[] ] := 
Module[ {k,S, edf, edb, W,G,GG,coor,alleds,col,edstab,edscol,opval},
k = SemigroupDegree[gen];
{S,edf,edb,W}=Closure[  OneSGT[k], Table[Evaluate[#**x]&,{x,gen}] ,
			Cayley->True,Witnesses->True, Full->True ];
 If[OptionValue[Full], Return[{S,edf,edb,W}]];
G = Graph[ TransitionsToEdges[edf,Labeled->False,Simple->False]];
coor = GraphEmbedding[G];

alleds = Sort@Join[ TransitionsToEdges[edf,Labeled->False,Simple->False],TransitionsToEdges[edb,Labeled->False,Simple->False]];
Echo[alleds];
If[(opval = OptionValue[GraphLayout])===Inherited,
GG = Graph[ alleds,VertexCoordinates->coor  ],
GG = Graph[ alleds,VertexCoordinates->opval  ]];
Assert[ k <= 6 ];
col = {Red,Blue,Green,Yellow,Cyan,Black};  (* FIX *)
If[ !OptionValue[Cayley],
HighlightGraph[ GG, G ]
(*else*),
edstab = Sort /@ GatherBy[ Join[edf,edb], Zwei ];
edstab = Apply[ DirectedEdge,Map[ Delete[#,2]& , edstab, {2} ],{2}];
edscol = Flatten@Table[   (#->col[[i]])&  /@ edstab[[i]], {i,3} ];
SetProperty[GG, EdgeStyle->edscol ]
]
];


(* ::Section::Closed:: *)
(*Rewrite*)


makeEpsilonEmpty[x_] := x /. _?EpsQ->"";

Options[SimplifyRWRules]={Full->True};
SimplifyRWRules[ eqns_List,  opts:OptionsPattern[] ] := 
Module[ {eqn,eq,lhs,oldeq},
	eqn = makeEpsilonEmpty[eqns];
    oldeq = eqn;
    While[ oldeq =!= {},
	eq  = First[oldeq];
	lhs = First[eq];
	Sow[eq];
	oldeq = Select[Rest@oldeq, (StringPosition[First[#],lhs]==={})&];
    ] // Reap // Last // Last
];


(* ::Section::Closed:: *)
(*Footer*)


Null
