(* ::Package:: *)

(* ::Title:: *)
(*Utils*)


(* ::Text:: *)
(*Author:	           Klaus Sutner*)
(*Affiliation: 	Carnegie Mellon University*)
(*email:       		sutner@cs.cmu.edu*)
(*url:         		http://www.cs.cmu.edu/~sutner*)
(*version:     	11.2*)


(* ::Section::Closed:: *)
(*Usage*)


BeginPackage["Automata11`utils`"];


SelectOptions::usage = "SelectOptions[cmd,opts]  selects those options from opts that make sense for cmd.";
AllOptions::usage = "AllOptions[sym] returns admissible types for the options associated with symbol sym.";
ResetOptions::usage = "ResetOptions[cmd] resets the options for cmd to default.";
ReplaceOption::usage = "ReplaceOption[cmd,op->val] sets the option op in cmd to val.";
GetOptionValue::usage = "GetOptionValue[cmd,opt] extracts the option value for command cmd, option opt.";
ParseRange::usage = "ParseRange[n]  processes a range description.";
SymbolList::usage = "SymbolList[sym,k]  generates a list of symbols of the form sym1, sym2, ..., symk.";
DomainFunction::usage = "DomainFunction[f] returns the domain of a function f.  Useful mostly when some definitions do not involve patterns.";
Drops::usage = "Drops is an option for DomainFunction.";
Curry::usage = "Curry[f] curries the function f.";
PositionOne::usage = "PositionOne[L,x]  evaluates to Position[L,x][[1,1]]; $Failed if position does not exist.";
PositionList::usage = "PositionList[A,B]  generates a flattened list of the positions of the elements of list B in list A.";
Pairs::usage = "Pairs[A,B] returns a list of pairs {ai,bi}.  Also works for k-tuples.";
FromBitVector::usage = "FromBitVector[S,U]  transforms bitvector S over universe U into a set.";
ToBitVector::usage = "ToBitVector[S,U]  transforms set S into a bitvector over universe U.";
AssignFunction::usage = "AssignFunction[f,A,B] assigns the function  f[ai] = bi to symbol f. A and B must be lists of equal length. Options Opts->{},Kary->False.";
UnorderedPairs::usage = "UnorderedPairs[n] produces all ordered pairs in [n].  With option Full->True includes diagonal. Also works on lists.";
PowerSet::usage = "PowerSet[A,k] generates all (k-) subsets of A.  Wrapper for Subsets.";
CartesianProduct::usage = "CartesianProduct[A,B] returns the Cartesian product of sets A and B.";
RankCartesian::usage = "RankCartesian[elm,cp]  computes the rank of the element elm of a Cartesian product given by the list cp of size.";
UnrankCartesian::usage = "UnrankCartesian[r,cp]  computes the the element of rank r of a Cartesian product given by the list cp of dimensions.";
RankingRules::usage = "RankingRules[L]  produces a list of rules that replace the elements of list L by their ranks.";
Shuffle::usage = "Shuffle[A,B]  shuffles two lists A and B; returns a list of lists.";
PerfectShuffle::usage = "PerfectShuffle[A,B]  computes the perfect shuffle of two lists A and B; must have the same length.";
ToClasses::usage = "ToClasses[L,R]  generates the equivalence classes of list L with respect to equivalence relation R. R can be given as a kernel function, Boolean relation, Boolean matrix or an index.";
Relation::usage = "Relation is an option for ToClasses.";
EquivalenceToIndex::usage = "EquivalenceToIndex[L,R]  computes an index of list L with respect to equivalence relation R (a Boolean function). ";
AnalyzeOrbit::usage = "AnalyzeOrbit[f,x] returns the transient and period of the orbit of x under function f.";
CycleDecomposition::usage = "CycleDecomposition[A,f] returns the cycle decomposition of A under permutation f.";
Closure::usage = "Closure[q0,{f,g,..},opts]  computes the closure of q0 under the functions f, g,..";
ClosureSearch::usage = "ClosureSearch[q0,x,{f,g,..},opts] tests whether x belongs to the closure of q0 under the functions f, g,..";
NonDecreasingSequences::usage = "NonDecreasingSequences[n,k] generates all non-decreasing sequences of length k over [n].";
TransMod::usage = "TransMod[x,t,p] computes the generalized mod of x wrto transient t and period m.";
SetClosure::usage = "SetClosure is an option for Closure.";
InitialSet::usage = "InitialSet is an option for Closure.";
Bound::usage = "Bound is an option for Closure.";
Rank::usage = "Rank is an option for Closure.";
Cayley::usage = "Cayley is an option for Closure.";
Witnesses::usage = "Witnesses is an option for Closure.";
Filter::usage = "Filter is an option for Closure.";
Index::usage = "Index is an option for Closure.";
Force::usage = "Force is an option for Closure.";



(* ::Subsection::Closed:: *)
(*Macros*)


Zwei = #[[2]]&;
Drei = #[[3]]&;   
FlattenOne = Flatten[#,1]&;
ToList= List@@#&;
Frequencies = Sort[Tally[#]] &;


Begin["`Private`"];


(* ::Section::Closed:: *)
(*System*)


(* ::Subsection::Closed:: *)
(*SelectOptions*)


(* ::Text:: *)
(*replace by FilterRules*)


PackageSymbol[SelectOptions,
"SelectOptions[cmd,opts]  selects those options from opts that make sense for cmd."];


ClearAll[SelectOptions]
SyntaxInformation[SelectOptions] = {"ArgumentsPattern"->{_,_,_.}};    
Options[SelectOptions] = {};
AllOptions[SelectOptions] = {};


SelectOptions[ cmd_Symbol, opts:OptionsPattern[] ] :=
    SelectOptions[ First /@ Options[cmd], opts ];
SelectOptions[ L_List, opts:OptionsPattern[] ] :=
    Sequence@@Select[ Flatten[{opts}], MemberQ[L, First[#]]& ];
SelectOptions[cmd_Symbol,opts___?OptionQ,{excl__}] :=
    SelectOptions[ Options[cmd],opts,{excl}];
SelectOptions[L_List,opts:OptionsPattern[],{excl__}] :=
    With[ {OO = Complement[First/@L,{excl}]},
        Sequence@@Select[ Flatten[{opts}],MemberQ[OO,First[#]]& ]
    ];


SelectOptions::args="Wrong argument(s), try SelectOptions[cmd,opt]";
SelectOptions[__]:=$Failed/;Message[SelectOptions::args];


(* ::Subsection::Closed:: *)
(*AllOptions*)


PackageSymbol[AllOptions,
"AllOptions[sym] returns admissible types for the options associated with symbol sym."];


ClearAll[AllOptions]
SyntaxInformation[AllOptions] = {"ArgumentsPattern"->{_,_.}};    
Options[AllOptions] = {};
AllOptions[AllOptions] = {};


AllOptions[cmd_,opt_] := Select[ AllOptions[cmd], (First[#]===opt &) ];


AllOptions::args="Wrong argument(s), try AllOptions[cmd]";
AllOptions[__]:=$Failed/;Message[AllOptions::args];


(* ::Subsection::Closed:: *)
(*ResetOptions*)


PackageSymbol[ResetOptions,
"ResetOptions[cmd] resets the options for cmd to default."];


ClearAll[ResetOptions]
ResetOptions::noassop = "Command `1` has no associated AllOptions information.";
SyntaxInformation[ResetOptions] = {"ArgumentsPattern"->{_}};    
Options[ResetOptions] = {};
AllOptions[ResetOptions] = {};


ResetOptions[cmd_] :=
    With[ {allopts = AllOptions[cmd]},
        If[ !MatchQ[allopts,{(_->{__})..}],
            Message[ResetOptions::noassop,cmd];
            Return[$Failed]
        ];
        Options[cmd] = Thread[First/@allopts->First/@Last/@allopts]
    ];


ResetOptions::args="Wrong argument(s), try ResetOptions[cmd]";
ResetOptions[__]:=$Failed/;Message[ResetOptions::args];


(* ::Subsection::Closed:: *)
(*ReplaceOption*)


PackageSymbol[ReplaceOption,
"ReplaceOption[cmd,op->val] sets the option op in cmd to val."];


ClearAll[ReplaceOption]
SyntaxInformation[ReplaceOption] = {"ArgumentsPattern"->{_,_}};    
Options[ReplaceOption] = {};
AllOptions[ReplaceOption] = {};


ReplaceOption[tt_, opt_ -> val_] := 
With[ {pos = Position[ tt, HoldPattern[opt -> _] ]},
	If[ pos === {}, Return[Append[tt, opt -> val]]];
	ReplacePart[ tt, pos[[1, 1]] -> (opt -> val) ]
];


ReplaceOption::args="Wrong argument(s), try ReplaceOption[cmd,opt]";
ReplaceOption[__]:=$Failed/;Message[ReplaceOption::args];


(* ::Subsection::Closed:: *)
(*GetOptionValue*)


PackageSymbol[GetOptionValue,
"GetOptionValue[cmd,opt] extracts the option value for command cmd, option opt."];


ClearAll[GetOptionValue]
SyntaxInformation[GetOptionValue] = {"ArgumentsPattern"->{_,_}};
Options[GetOptionValue] = {};
AllOptions[GetOptionValue] = {};


GetOptionValue[term_] := Cases[  List@@term, _Rule ];
GetOptionValue[ obj_, opt_ ] := opt /. Cases[  List@@obj, _Rule ] /. Options[Head[obj]];


GetOptionValue::args="Wrong argument(s), try GetOption[cmd]";
GetOptionValue[__]:=$Failed/;Message[GetOption::args];


(* ::Subsection::Closed:: *)
(*ParseRange*)


(* ::Text:: *)
(*standard convention*)
(*n ->     {0,n}*)
(*{n} ->  {n,n}*)
(*{n,m} *)


PackageSymbol[ParseRange,
"ParseRange[n]  processes a range description."];


SyntaxInformation[ParseRange] = {"ArgumentsPattern"->{__}};    
Options[ParseRange] = {};
AllOptions[ParseRange] = {};


ParseRange[n_Integer?NonNegative] := {0,n};
ParseRange[{n_Integer?NonNegative}] := {n,n};
ParseRange[{n1_Integer?NonNegative,n2_Integer?NonNegative}] := {n1,n2}  /;   n1 <= n2;


ParseRange::args = "Wrong argument(s), try n, {n}, {n,m}.";
ParseRange[_] :=$Failed/;Message[ParseRange::args];


(* ::Section::Closed:: *)
(*Core*)


(* ::Subsection::Closed:: *)
(*SymbolList*)


PackageSymbol[SymbolList,
"SymbolList[sym,k]  generates a list of symbols of the form sym1, sym2, ..., symk."];


SyntaxInformation[SymbolList] = {"ArgumentsPattern"->{_,_}};    
Options[SymbolList] = {};
AllOptions[SymbolList] = {};


SymbolList[ str_String, ll_Integer ] := SymbolList[ str, Range[ll] ]; 
SymbolList[ str_String, LL_List] := ToExpression /@ (StringJoin[ str, ToString[#] ]& /@ LL);


SymbolList::args="Wrong argument(s), try SymbolList[sym,n]";
SymbolList[__]:=$Failed/;Message[SymbolList::args];


(* ::Subsection::Closed:: *)
(*DomainFunction*)


PackageSymbol[DomainFunction,
"DomainFunction[f] returns the domain of a function f.  Useful mostly when some definitions do not involve patterns."];


PackageSymbol[Drops,
"Drops is an option for DomainFunction."];


SyntaxInformation[DomainFunction] = {"ArgumentsPattern" -> {_, OptionsPattern[]}};
Options[DomainFunction] = {Drops->0};
AllOptions[DomainFunction] = {};


DomainFunction[ff_, opts:OptionsPattern[]] :=
    With[ {d = OptionValue[Drops]},
        First /@ ((First /@ Drop[DownValues[ff],-d]) /. ff -> List) ];


DomainFunction::args="Wrong argument(s), try DomainFunction[f]";
DomainFunction[__]:=$Failed/;Message[DomainFunction::args];




(* ::Subsection::Closed:: *)
(*Curry*)


PackageSymbol[Curry,
"Curry[f] curries the function f."];


SyntaxInformation[Curry] = {"ArgumentsPattern"->{_}};    

Options[Curry] = {};

AllOptions[Curry] = {};


Curry[fff_] := Function[yyy,Function[xxx,fff[xxx,yyy]]];


Curry::args="Wrong argument(s), try Curry[f]";
Curry[__]:=$Failed/;Message[Curry::args];


(* ::Subsection::Closed:: *)
(*PositionOne*)


PackageSymbol[PositionOne,
"PositionOne[L,x]  evaluates to Position[L,x][[1,1]]; $Failed if position does not exist."];


SyntaxInformation[PositionOne] = {"ArgumentsPattern"->{_,_}};    

Options[PositionOne] = {};

AllOptions[PositionOne] = {};


PositionOne[L_,x__] := 
     With[ {pos = Position[L,x]}, If[ pos==={}, $Failed, pos[[1,1]] ] ];


PositionOne::args="Wrong argument(s), try PositionOne[L,x]";
PositionOne[__]:=$Failed/;Message[PositionOne::args];


(* ::Subsection::Closed:: *)
(*PositionList*)


PackageSymbol[PositionList,
"PositionList[A,B]  generates a flattened list of the positions of the elements of list B in list A."];


SyntaxInformation[PositionList] = {"ArgumentsPattern"->{_,_}};    

Options[PositionList] = {};

AllOptions[PositionList] = {};


PositionList[L1_List,L2_List] := Flatten[ Position[L1,#]& /@ L2 ];
PositionList[L1_List,L2_ ] := PositionList[L1,{L2}];


PositionList::args="Wrong argument(s), try PositionList[L,K]";
PositionList[__]:=$Failed/;Message[PositionList::args];


(* ::Subsection::Closed:: *)
(*Pairs*)


PackageSymbol[Pairs,
"Pairs[A,B] returns a list of pairs {ai,bi}.  Also works for k-tuples."];


SyntaxInformation[Pairs] = {"ArgumentsPattern"->{_,__}};    

Options[Pairs] = {};

AllOptions[Pairs] = {};


Pairs[LL__List] := Thread[{LL}];


Pairs::args="Wrong argument(s), try Pairs[A,B]";
Pairs[_]:=$Failed/;Message[Pairs::args];


(* ::Subsection::Closed:: *)
(*FromBitVector*)


PackageSymbol[FromBitVector,
"FromBitVector[S,U]  transforms bitvector S over universe U into a set."];


SyntaxInformation[FromBitVector] = {"ArgumentsPattern" -> {_,_}};

Options[FromBitVector] = {};

AllOptions[FromBitVector] = {};


FromBitVector[B_List,UU_List] := Pick[ UU, B, 1 ];
FromBitVector[B_List] :=  Pick[ Range[Length[B]], B, 1 ];


FromBitVector::args="Wrong argument(s), try FromBitVector[B,U]";
FromBitVector[__]:=$Failed/;Message[FromBitVector::args];


(* ::Subsection::Closed:: *)
(*ToBitVector*)


PackageSymbol[ToBitVector,
"ToBitVector[S,U]  transforms set S into a bitvector over universe U."];


SyntaxInformation[ToBitVector] = {"ArgumentsPattern" -> {_,_}};

Options[ToBitVector] = {};

AllOptions[ToBitVector] = {};


ToBitVector[S_List,UU_List]   := If[MemberQ[S,#],1,0]& /@ UU;
ToBitVector[S_List,n_Integer] := If[MemberQ[S,#],1,0]& /@ Range[n];


ToBitVector::args="Wrong argument(s), try ToBitVector[S,U]";
ToBitVector[__]:=$Failed/;Message[ToBitVector::args];


(* ::Subsection::Closed:: *)
(*AssignFunction*)


PackageSymbol[AssignFunction,
"AssignFunction[f,A,B] assigns the function  f[ai] = bi to symbol f. A and B must be lists of equal length. Options Opts->{},Kary->False."];


SyntaxInformation[AssignFunction] = {"ArgumentsPattern"->{_,_,_.,OptionsPattern[]}};
Attributes[AssignFunction] = {HoldFirst};
Options[AssignFunction] = {Opts->{},Kary->False};
AllOptions[AssignFunction] = {Opts->{{},Options},Kary->{False,True}};


AssignFunction[h_,dom_List,cod_List,opts:OptionsPattern[]] :=
    AssignFunction[h,Pairs[dom,cod],opts]   /;  Length[dom]==Length[cod];

AssignFunction[h_,L_List, opts:OptionsPattern[] ] :=
    (
        ClearAll[h];
        If[ OptionValue[Kary],
            Scan[(h[Sequence@@First[#1]] = Last[#1])&,L],
            Scan[(h[First[#1]] = Last[#1])&,L]
        ];
        SetAttributes[h,OptionValue[Opts]]
    );


(* ::Input:: *)
(*Test[*)
(*Module[{f},*)
(*AssignFunction[f,{{1,a},{2,b},{3,c}},Opts->Listable];*)
(*f[{1,2,3}]],*)
(*{a,b,c}*)
(*];*)
(**)


(* ::Input:: *)
(*Test[*)
(*Module[{f},*)
(*AssignFunction[ f, CartesianProduct[2,2], Range[4], Kary->True ];*)
(*f[1,2]],*)
(*2*)
(*];*)


AssignFunction::args="Wrong argument(s), try AssignFunction[f,A,B]";
AssignFunction[__]:=$Failed/;Message[AssignFunction::args];


(* ::Section::Closed:: *)
(*Combinatorics*)


(* ::Subsection::Closed:: *)
(*Tuples*)


Unprotect[Tuples];
Tuples[k_Integer?Positive,n_Integer]:= Tuples[Range[k],n];
Tuples[k_Integer?Negative,n_Integer]:= Tuples[Range[0,-k-1],n];
Tuples[L_List,n_Integer?Negative]:= FlattenOne[Table[Tuples[L,{k}],{k,0,-n}]];
Protect[Tuples];


(* ::Subsection::Closed:: *)
(*UnorderedPairs*)


PackageSymbol[UnorderedPairs,
"UnorderedPairs[n] produces all ordered pairs in [n].  With option Full->True includes diagonal. Also works on lists."];

SyntaxInformation[UnorderedPairs] = {"ArgumentsPattern"->{_,OptionsPattern[]}};    

Options[UnorderedPairs] ={Full->False};

AllOptions[UnorderedPairs] = {Full->{False,True}};


UnorderedPairs[ n_Integer?Positive ,opts:OptionsPattern[] ] := 
    If[  OptionValue[Full],  
               Sort[ Join[ Pairs[Range[n],Range[n]], Subsets[Range[n],{2}] ] ], 
               Subsets[Range[n],{2}]
    ];

(* probably should not use Part *) 
UnorderedPairs[ L_List, opts:OptionsPattern[] ] := L[[#]]& /@  UnorderedPairs[Length[L],opts ];


UnorderedPairs::args="Wrong argument(s), try UnorderedPairs[n]";
UnorderedPairs[__]:=$Failed/;Message[UnorderedPairs::args];


(* ::Subsection::Closed:: *)
(*PowerSet*)


PackageSymbol[PowerSet,
"PowerSet[A,k] generates all (k-) subsets of A.  Wrapper for Subsets."];


SyntaxInformation[PowerSet] = {"ArgumentsPattern"->{_,_.}};    

Options[PowerSet] = {};

AllOptions[PowerSet] = {};


PowerSet[ L_List ] := Subsets[L];
PowerSet[ L_List, k_Integer?Positive ] := Subsets[L,{k}];
PowerSet[ L_List, k_Integer?Negative ] := Subsets[L,-k];
PowerSet[ r_Integer?Positive, k_Integer ] := PowerSet[Range[r],k];


PowerSet::args="Wrong argument(s), try PowerSet[A,k]";
PowerSet[__]:=$Failed/;Message[PowerSet::args];


(* ::Subsection::Closed:: *)
(*CartesianProduct*)


PackageSymbol[CartesianProduct,
"CartesianProduct[A,B] returns the Cartesian product of sets A and B."];


SyntaxInformation[CartesianProduct] = {"ArgumentsPattern" -> {___}};
Options[CartesianProduct] = {};
AllOptions[CartesianProduct] = {};


CartesianProduct[___,{},___] = {};
CartesianProduct[ LL___List ] := Tuples[{LL}];
CartesianProduct[ ints__Integer ] := Tuples[Range /@ {ints}];


CartesianProduct::args="Wrong argument(s), try CartesianProduct[A,B]";
CartesianProduct[__]:=$Failed/;Message[CartesianProduct::args];


(* ::Subsection::Closed:: *)
(*RankCartesian*)


PackageSymbol[RankCartesian,
"RankCartesian[elm,cp]  computes the rank of the element elm of a Cartesian product \
given by the list cp of size."];


(* ::Text:: *)
(*Cartesian products are 0/1-indexed  (Subscript[n, k])\[Cross](Subscript[n, k-1])\[Cross]\[Ellipsis]\[Cross](Subscript[n, 1]), rank is likewise 0-indexed *)
(*should probably be more flexible;  0-indexed rank:*)
(*rk({Subscript[a, k],...,Subscript[a, 1]}) = \!\( *)
(*\*SubsuperscriptBox[\(\[Sum]\), \(i = 1\), \(k\)]\( *)
(*\*SubscriptBox[\(a\), \(i\)] *)
(*\*SubscriptBox[\(r\), \(i\)]\)\)  where  Subscript[r, i ]= \!\( *)
(*\*SubscriptBox[\(\[Product]\), \(j < i\)]*)
(*\*SubscriptBox[\(n\), \(j\)]\).*)


ClearAll[RankCartesian]
SyntaxInformation[RankCartesian] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};    
Options[RankCartesian]={ZeroIndexed->True};
AllOptions[RankCartesian] = {};


RankCartesian[ elm_List, cp_List, opts:OptionsPattern[]] := 
With[ {rks = FoldList[  #1 #2&, 1,  Reverse@Rest[cp] ]}, 
	If[ OptionValue[ZeroIndexed], 
		rks . Reverse[elm],  rks . Reverse[elm-1]+1 ] ]/; Length[elm]==Length[cp];


(* ::Input:: *)
(*RankCartesian[ #, {3,5} ]& /@ (CartesianProduct[3,5]-1)*)


(* ::Input:: *)
(*RankCartesian[ #, {3,5}, ZeroIndexed->False ]& /@ CartesianProduct[3,5]*)


(* ::Input:: *)
(*(RankCartesian[ #, {3,5} ]& /@ (CartesianProduct[3,5]))*)


(* ::Subsection::Closed:: *)
(*UnrankCartesian*)


PackageSymbol[UnrankCartesian,
"UnrankCartesian[r,cp]  computes the the element of rank r of a Cartesian product \
given by the list cp of dimensions."];


(* ::Text:: *)
(*Cartesian products are 0-indexed  (Subscript[n, k])\[Cross](Subscript[n, k-1])\[Cross]\[Ellipsis]\[Cross](Subscript[n, 1]), rank is 0-indexed*)
(*rk({Subscript[a, k],...,Subscript[a, 1]}) = \!\( *)
(*\*SubsuperscriptBox[\(\[Sum]\), \(i = 1\), \(k\)]\( *)
(*\*SubscriptBox[\(a\), \(i\)] *)
(*\*SubscriptBox[\(r\), \(i\)]\)\)  where  Subscript[r, i ]= \!\( *)
(*\*SubscriptBox[\(\[Product]\), \(j < i\)]*)
(*\*SubscriptBox[\(n\), \(j\)]\).*)


ClearAll[UnrankCartesian]
SyntaxInformation[UnrankCartesian] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};    
Options[UnrankCartesian]={ZeroIndexed->True};
AllOptions[UnrankCartesian] = {};


UnrankCartesian[ r_Integer, cp_List, opts:OptionsPattern[] ] := 
If[ OptionValue[ZeroIndexed],
Reverse@First@Last[Reap[ Fold[ (Sow[Mod[#1,#2]]; Quotient[#1,#2])&, r,  Reverse[cp] ] ]],
Reverse@First@Last[Reap[ Fold[ (Sow[Mod[#1,#2]]; Quotient[#1,#2])&, r-1,  Reverse[cp] ] ]]+1 ];


(* ::Input:: *)
(*UnrankCartesian[ #, {3,5} ]& /@ Range[0,14]*)


(* ::Input:: *)
(*UnrankCartesian[ #, {3,5}, ZeroIndexed->False ]& /@ Range[15]*)


(* ::Subsection::Closed:: *)
(*RankingRules*)


PackageSymbol[RankingRules,
"RankingRules[L]  produces a list of rules that replace the elements of list L by their ranks."];


SyntaxInformation[RankingRules] = {"ArgumentsPattern"->{_,OptionsPattern[]}};    

Options[RankingRules]={Direction->Forward,Index->1};

AllOptions[RankingRules] = {};


RankingRules[L_List, opts:OptionsPattern[] ] := 
With[ { ind = OptionValue[Index]},
    	If[ OptionValue[Direction] === Forward, 
      		Thread[ L ->  Range[ind, ind + Length[L] - 1] ],
      		Thread[ Range[ind, ind + Length[L] - 1] ->  L ] ]
];


RankingRules::args="Wrong argument(s), try RankingRules[L]";
RankingRules[__]:=$Failed/;Message[RankingRules::args];


(* ::Subsection::Closed:: *)
(*Shuffle*)


PackageSymbol[Shuffle,
"Shuffle[A,B]  shuffles two lists A and B; returns a list of lists."];


SyntaxInformation[Shuffle] = {"ArgumentsPattern"->{_,_}};    

Options[Shuffle] = {};

AllOptions[Shuffle] = {};


Shuffle[{},L_List] = {L};
Shuffle[L_List,{}] = {L};
Shuffle[{a_,aa___},{b_,bb___}] := 
     Union[Prepend[#,a]& /@ Shuffle[{aa},{b,bb}], Prepend[#,b]& /@ Shuffle[{a,aa},{bb}]];
Shuffle[n1_Integer,n2_Integer] := Shuffle[ Table[1,{n1}], Table[2,{n2}] ];


Shuffle::args="Wrong argument(s), try Shuffle[L,K]";
Shuffle[__]:=$Failed/;Message[Shuffle::args];


(* ::Subsection::Closed:: *)
(*PerfectShuffle*)


PackageSymbol[PerfectShuffle,
"PerfectShuffle[A,B]  computes the perfect shuffle of two lists A and B; must have the same length."];

SyntaxInformation[PerfectShuffle] = {"ArgumentsPattern" -> {_,_}};

SyntaxInformation[PerfectShuffle] = {"ArgumentsPattern"->{_,_}};    

Options[PerfectShuffle] = {};

AllOptions[PerfectShuffle] = {};


PerfectShuffle[L_List,K_List] := Flatten@Thread[{L,K}] /;  Length[L]==Length[K];


PerfectShuffle::args="Wrong argument(s), try PerfectShuffle[A,B]";
PerfectShuffle[__]:=$Failed/;Message[PerfectShuffle::args];


(* ::Subsection::Closed:: *)
(*ToClasses*)


(* ::Text:: *)
(*index is the canonical selecter function (as a list)*)


PackageSymbol[ToClasses,
"ToClasses[L,R]  generates the equivalence classes of list L with respect to equivalence relation R. \
R can be given as a kernel function, Boolean relation, Boolean matrix or an index."];


SyntaxInformation[ToClasses] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};    
Options[ToClasses]={Type->Relation};
AllOptions[ToClasses] = {Type->{Function,Relation,Matrix,Index}};


ToClasses[ A_List, r_, opts:OptionsPattern[]]:=
    Switch[  OptionValue[Type],
      Function,       GatherBy[ A, r ]    (* Reap[Scan[Sow[#,{r[#]}]& , A]][[2]]  *),
      Relation,       Reap[Scan[Sow[#[[1]],{#[[2]]}]& ,Pairs[A,EquivalenceToIndex[A,r]]]][[2]],
      Matrix,         Union[(A[[Flatten[Position[#1,1]]]]&)/@ r],
      Index,          Pick[ A, r, # ]& /@  Union[r]    (* Reap[Scan[Sow[#[[1]],{#[[2]]}]& ,Pairs[A,r]]][[2]] *)
    ];


ToClasses::args="Wrong argument(s), try ToClasses[A,R]";
ToClasses[__]:=$Failed/;Message[ToClasses::args];


(* ::Subsection::Closed:: *)
(*EquivalenceToIndex*)


(* ::Text:: *)
(*add option for number of classes?*)
(*ranking slightly faster*)


PackageSymbol[EquivalenceToIndex,
"EquivalenceToIndex[L,R]  computes an index of list L with respect to equivalence relation R (a Boolean function). "];


ClearAll[EquivalenceToIndex]
SyntaxInformation[EquivalenceToIndex] = {"ArgumentsPattern"->{_,_.,OptionsPattern[]}};    
Options[EquivalenceToIndex] = {Normal->True,Full->False,Type->Relation};
AllOptions[EquivalenceToIndex] = {};


EquivalenceToIndex[ L_List, opts:OptionsPattern[] ] := 
Module[ {hh,top,findpos,LL},
	hh[_] = 0;
	If[ OptionValue[Normal],    (* ranking kernel [n] --> [r] *) 
		top = 1;
		hh[First[L]] = 1;
		If[ hh[#]==0, hh[#]=++top, hh[#] ]& /@ L,
	(* else canonical  [n] --> [n]  *)
		findpos[{a_,b_}] := If[ hh[b]==0, hh[b]=a, hh[b] ];    
		LL = Pairs[ Range[Length[L]], L ];
		hh[ LL[[1]] ] = 1;
		findpos /@ LL
	]
];


 
EquivalenceToIndex[ L_List, R_, opts:OptionsPattern[] ] := 
Module[ {h,hh,findindex,top},
    top = 1;
    h[_] := 0;
    h[First[L]] = 1;
    hh[1] = First[L];
        
    findindex[x_] := Catch[
       Scan[ If[ R[x,hh[#]], Throw[ h[x]=# ] ]&, Range[top] ];
             h[x]=++top; hh[top]=x ; Throw[top] ];    
        
    findindex /@ L
];


EquivalenceToIndex::args="Wrong argument(s), try EquivalenceToIndex[A,R]";
EquivalenceToIndex[__]:=$Failed/;Message[EquivalenceToIndex::args];


(* ::Subsection::Closed:: *)
(*AnalyzeOrbit*)


PackageSymbol[AnalyzeOrbit,
"AnalyzeOrbit[f,x] returns the transient and period of the orbit of x under function f."];

$boundAnalyzeOrbit = 5000;


AnalyzeOrbit::trunc = "Search for anchor point was truncated after `1` steps; you might consider to increase the value of option Bound.";

SyntaxInformation[AnalyzeOrbit] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};

Options[AnalyzeOrbit] = {Cycle->None,Periodic->False,Bound:>$boundAnalyzeOrbit,
Mode->Teleport};

AllOptions[AnalyzeOrbit] = {Cycle->{None,True,False},Periodic->{False,True},
Bound->{$boundAnalyzeOrbit,_Integer},Mode->{Teleport,None}};


AnalyzeOrbit[ ff_, aa_, opts:OptionsPattern[] ] :=
    Module[ {slow,fast,per,trns,bnd,trnc,gg},
        trnc = OptionValue[Bound];
        bnd = 0;
        gg = ff[ff[#]]&;
        If[ OptionValue[Periodic], 
			Return[ analyzePeriodicOrbit[ ff, aa, opts ] ] ];
        If[ None === OptionValue[Mode],

        (*  find a point slow on the cycle  *)
            slow = ff[aa];
            fast = gg[aa];
            While[ slow=!=fast && ++bnd < trnc, 
                slow = ff[slow];
                fast = gg[fast];
            ];

            If[ bnd == trnc,
                Message[AnalyzeOrbit::trunc,trnc];
                Return[$Failed]
            ];

            (*  compute the length per of the cycle  *)
            per = 1;
            NestWhile[ (per++; ff[#])&, ff[slow], #=!=slow& ];

         ,(* else: teleport *)

            slow = aa;
            fast = ff[aa];
            per = bnd = 1;
            While[ slow != fast && bnd < trnc,
                If[ per == bnd,
                    slow = fast;
                    per = 0;
                    bnd *= 2
                ];
                fast = ff[fast];
                per++ 
            ];
            If[ bnd >= trnc,
                Message[AnalyzeOrbit::trunc,trnc];
                Return[$Failed]
            ];
        ];

    (*  compute the transient length  *)
        trns = 0;
        slow = First@NestWhile[ (trns++; {ff[#[[1]]],ff[#[[2]]]})&,
					{aa,Nest[ff,aa,per]}, #[[1]]=!=#[[2]]& ];
        Switch[ OptionValue[Cycle],
                None,    {trns,per},
                False,   {trns,per,slow},
                True,    {trns,per,NestList[ff,slow,per-1]} 
            ]
    ];


Options[analyzePeriodicOrbit] = Options[AnalyzeOrbit];
analyzePeriodicOrbit[ ff_, aa_, opts:OptionsPattern[] ] :=
    Module[ {pcnt}, 

        (*  orbit must be a cycle, does not check  *)
        pcnt = 1;
        NestWhile[ (pcnt++;
                    ff[#])&, ff[aa], #=!=aa& ];
        Switch[ OptionValue[Cycle],
            None,       {0,pcnt},
            False,      {0,pcnt,aa},
            True,       {0,pcnt,NestList[ff,aa,pcnt-1]} 
        ]
    ];


AnalyzeOrbit::args="Wrong argument(s), try AnalyzeOrbit[f,a]";
AnalyzeOrbit[__]:=$Failed/;Message[AnalyzeOrbit::args];


(* ::Subsection::Closed:: *)
(*CycleDecomposition*)


PackageSymbol[CycleDecomposition,
"CycleDecomposition[A,f] returns the cycle decomposition of A under permutation f."];


SyntaxInformation[CycleDecomposition] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};
Options[CycleDecomposition] = {Full->False,Bound:>$cycleDecompBound};
AllOptions[CycleDecomposition] = {Full->{False,True}};

$cycleDecompBound = 100;


CycleDecomposition[ A_List, ff_, opts:OptionsPattern[] ] := 
Module[ {n,cyccnt,lencnt,mark,full,i,res,x,x0,cnt},
	n = Length[A];
	cyccnt = 0;
	lencnt[_]  := 0;
	mark[_] := False;
	full = OptionValue[Full];

	i = 1;
	res = Reap[While[  i <= n,
		While[ i<= n && mark[A[[i]]], i++ ];
		If[ i > n,Break[] ];
		x0 = x = A[[i]];
		mark[x0] = True;
		cnt = 1;
		cyccnt++;
		If[ full, Sow[{x,cyccnt}] ];
		While[  ff[x] != x0,  
			cnt++; x = ff[x]; mark[x]=True;
			If[ full, Sow[{x,cyccnt}] ]
		];
		lencnt[cnt]++;
	]];
	If[ full,  
			Map[ First, GatherBy[ res[[-1,-1]], Last ], {2} ],
			Sort@auxDownValuesToPairs[lencnt]
	]
];

Clear[auxDownValuesToPairs]
auxDownValuesToPairs[ff_] := 
	Flatten/@ (Most[DownValues[ff]] /. {ff->List, HoldPattern->Identity,RuleDelayed-> List});


CycleDecomposition::args="Wrong argument(s), try CycleDecomposition[A,f]";
CycleDecomposition[__]:=$Failed/;Message[CycleDecomposition::args];


(* ::Subsection::Closed:: *)
(*Closure*)


(* ::Text:: *)
(*Returns $Failed only in auxGenerateClosure. *)
(*In main function, bound is on number of application of operators, not on number of generated objects. *)
(*Replace by NestWhile.*)
(*SetClosure     f : A \[LongRightArrow] A   versus  f : A \[LongRightArrow] \[GothicCapitalP](A)*)
(*InitialSet          a \[Element] A  versus  a \[SubsetEqual] A*)


PackageSymbol[Closure,
"Closure[q0,{f,g,..},opts]  computes the closure of q0 under the functions f, g,.."];


SyntaxInformation[Closure] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};    
$boundClosure = 5000;
Options[Closure] = {SetClosure -> False, InitialSet -> False,
   Bound :> $boundClosure, Rank -> False, Cayley -> False, Witnesses->False, Full->False,
   Filter -> Identity, Index -> 1, SameTest -> SameQ, 
   Force -> False };
Options[auxGenerateClosure] = Options[Closure];
AllOptions[Closure] = {};


Closure[ q0_, ff_List, opts:OptionsPattern[] ] :=
    Module[ {Q0, S, sc, cb, rk, SS, cnt,same},
    	same = OptionValue[SameTest];
        Q0 = If[ OptionValue[InitialSet], q0, {q0} ];
        sc = If[ OptionValue[SetClosure], 2, 1 ];
        cb = OptionValue[Bound];
        S = {};
        If[ OptionValue[Cayley],
			Return[ auxGenerateClosure[ Q0, ff, sc, opts ] ] ];
        
        (* FP: the set of active elements is empty *)
        If[ OptionValue[Rank],
            cnt = 0;
            Scan[(rk[#] = cnt) &, Q0 ];
            FixedPoint[
                 (  S = Union[S, #, SameTest->same ];
                    SS = Complement[Flatten[Function[z, z /@ #] /@ ff, sc], S, SameTest->same ];
                    cnt++;
                    Scan[(rk[#] = cnt) &, SS];
                    SS) &,
                 Q0, cb ];
            {S, rk /@ S},
        (*else*)
            FixedPoint[
             ( S = Union[S, #, SameTest->same ];
               Complement[Flatten[Function[z, z /@ #] /@ ff, sc], S, SameTest->same ]) &,
             Q0, cb ];
            S
        ]
    ];

Closure[ q0_, ff__, opts:OptionsPattern[] ] := Closure[ q0, {ff}, opts ];

auxGenerateClosure[ Q0_List, F_List, sc_Integer, opts:OptionsPattern[] ] :=
Module[ {smax, filt, rtyp, k, S, h, cnt, rpf, rpb, A, x, a, act, wit, trf, trb, res},
    smax = OptionValue[Bound];
    filt = OptionValue[Filter];
    cnt = Length[Q0];
    k = Length[F];
    S = act = Q0;
    Scan[ (wit[#]={})&, S ];
    AssignFunction[h, Q0, Range[cnt]];  (* currently too general *)
    rpb = Reap[ rpf = Reap[
      While[ act =!= {} && Length[S] < smax,
           {x, act} = {First@act, Rest@act};
           Do[ 
                A = F[[i]][x];
                If[ sc == 1,
                    A = {A}
                ];
                Scan[  
                     ( a = filt[#];
                       If[ !IntegerQ[h[a]],     (* forward edge, MemberQ too slow *)
                           AppendTo[S, a];
                           AppendTo[act, a];
                           h[a] = ++cnt;
                           Sow[{h[x],i,h[a]},"forw"];
                           wit[a] = Append[ wit[x], i ],
                       (* else backward *)
                           Sow[ {h[x],i,h[a]}, "back" ] 
                       ];
                      ) &, A ],
                {i, k} 
            ];
       ],"forw"];,"back"];

	If[ Length[S] >= smax && !OptionValue[Force], 
        Return[ $Failed ] ];

    trf = If[ rpf[[2]]=!={}, rpf[[2,1]], {} ];
    trb = If[ rpb[[2]]=!={}, rpb[[2,1]], {} ];

    res = If[ OptionValue[Full], 
               {S,trf,trb}, 
               {S,Join[trf,trb]} ];

    If[ OptionValue[Witnesses],
        Append[ res, wit/@S ],
        res
    ]
];


Closure::args="Wrong argument(s), try Closure[a,{f,g}]";
Closure[__]:=$Failed/;Message[Closure::args];


(* ::Subsection::Closed:: *)
(*ClosureSearch*)


PackageSymbol[ClosureSearch,
"ClosureSearch[q0,x,{f,g,..},opts] tests whether x belongs to the closure of q0 under the functions f, g,.."];


$boundClosureSearch = 5000;

ClosureSearch::bnd  = "ClosureSearch has reached its search bound; try to increase option Bound.";
 
SyntaxInformation[ClosureSearch] = {"ArgumentsPattern"->{_,_,__,OptionsPattern[]}};    

Options[ClosureSearch] = {SetClosure -> False, InitialSet -> False, Bound :> $boundClosureSearch};

AllOptions[ClosureSearch] = {};

ClosureSearch[ q0_, targ_, ff_List, opts:OptionsPattern[]] :=
    Module[ {act, S, sc, cb},
        act = If[ OptionValue[InitialSet],
                  q0,
                  {q0}
              ];
        sc = If[ OptionValue[SetClosure],
                 2,
                 1
             ];
        cb = OptionValue[Bound];
        S = {};
        While[ Length[act] > 0 && Length[S] < cb,
             S   = Union[S, act];
             If[ MemberQ[S, targ],
                 Return[True]
             ];
             act = Complement[Flatten[ Function[z, z /@ act] /@ ff, sc], S];
         ];
        If[ Length[S] >= cb,
            Message[ ClosureSearch::bnd ];
            Return[$Failed]
        ];
        False
    ];

ClosureSearch[ q0_, targ_, ff__, opts:OptionsPattern[]] :=
               ClosureSearch[ q0, targ, {ff}, opts ];

ClosureSearch::args = "Wrong argument(s), try ClosureSearch[a,f]";
ClosureSearch[__]:=$Failed/;Message[ClosureSearch::args];


(* ::Subsection::Closed:: *)
(*NonDecreasingSequences*)


PackageSymbol[NonDecreasingSequences,
"NonDecreasingSequences[n,k] generates all non-decreasing sequences of length k over [n]."];


ClearAll[NonDecreasingSequences]
SyntaxInformation[NonDecreasingSequences] = {"ArgumentsPattern"->{_,_,OptionsPattern[]}};
Options[NonDecreasingSequences]={Full->True};
AllOptions[NonDecreasingSequences]={};


NonDecreasingSequences[n_Integer?Positive, 0, opts:OptionsPattern[] ] := {{}};
NonDecreasingSequences[n_Integer?Positive, k_Integer?Positive, opts:OptionsPattern[] ] :=
Module[ {ind, iter},
	ind = Table[Unique[], {k}];
	If[ OptionValue[Full],
		iter = Thread[{ind, Prepend[Most[ind], 1], Table[n, {k}]}],
		iter = Thread[{ind, Prepend[Most[ind], 0]+1, Table[n, {k}]}]
	];
	Flatten[ Table[Evaluate[ind], Evaluate[Sequence @@ iter]], k-1]
];
NonDecreasingSequences[n_Integer?Positive, k_Integer?Negative, opts:OptionsPattern[] ] :=
FlattenOne@Table[ NonDecreasingSequences[n,i,opts], {i,0,-k} ];
NonDecreasingSequences[L_List,k_Integer?Positive,opts___?OptionQ] := 
	L[[#]]& /@  NonDecreasingSequences[Length[L],k,opts];


(* ::Subsection::Closed:: *)
(*TransMod*)


PackageSymbol[TransMod,
"TransMod[x,t,p] computes the generalized mod of x wrto transient t and period m."];


ClearAll[TransMod]
SyntaxInformation[TransMod] = {"ArgumentsPattern"->{_,_,_.,OptionsPattern[]}};
Options[TransMod]={Index->0};
AllOptions[TransMod]={};


TransMod[ x_Integer, {t_Integer?NonNegative, p_Integer?Positive}, opts:OptionsPattern[] ] :=
TransMod[ x, t, p, opts ];

TransMod[ x_Integer, t_Integer?NonNegative, p_Integer?Positive, opts:OptionsPattern[] ] :=
With[ { tt = t + OptionValue[Index] },
      	If[x < tt, x, Mod[x - tt, p] + tt]
];


(* ::Section::Closed:: *)
(*Footer*)


(* ::Text:: *)
(*Ultimately  protect stuff here. *)


SelectOptions::args = "Wrong argument(s), try SelectOptions[cmd,opts]";
SelectOptions[__] := $Failed /; Message[SelectOptions::args];
AllOptions::args = "Wrong argument(s), try AllOptions[sym]";
AllOptions[__] := $Failed /; Message[AllOptions::args];
ResetOptions::args = "Wrong argument(s), try ResetOptions[cmd]";
ResetOptions[__] := $Failed /; Message[ResetOptions::args];
ReplaceOption::args = "Wrong argument(s), try ReplaceOption[cmd,op->val]";
ReplaceOption[__] := $Failed /; Message[ReplaceOption::args];
GetOptionValue::args = "Wrong argument(s), try GetOptionValue[cmd,opt]";
GetOptionValue[__] := $Failed /; Message[GetOptionValue::args];
ParseRange::args = "Wrong argument(s), try ParseRange[n]";
ParseRange[__] := $Failed /; Message[ParseRange::args];
SymbolList::args = "Wrong argument(s), try SymbolList[sym,k]";
SymbolList[__] := $Failed /; Message[SymbolList::args];
DomainFunction::args = "Wrong argument(s), try DomainFunction[f]";
DomainFunction[__] := $Failed /; Message[DomainFunction::args];
Drops::args = "Wrong argument(s), try Drops";
Drops[__] := $Failed /; Message[Drops::args];
Curry::args = "Wrong argument(s), try Curry[f]";
Curry[__] := $Failed /; Message[Curry::args];
PositionOne::args = "Wrong argument(s), try PositionOne[L,x]";
PositionOne[__] := $Failed /; Message[PositionOne::args];
PositionList::args = "Wrong argument(s), try PositionList[A,B]";
PositionList[__] := $Failed /; Message[PositionList::args];
Pairs::args = "Wrong argument(s), try Pairs[A,B]";
Pairs[__] := $Failed /; Message[Pairs::args];
FromBitVector::args = "Wrong argument(s), try FromBitVector[S,U]";
FromBitVector[__] := $Failed /; Message[FromBitVector::args];
ToBitVector::args = "Wrong argument(s), try ToBitVector[S,U]";
ToBitVector[__] := $Failed /; Message[ToBitVector::args];
AssignFunction::args = "Wrong argument(s), try AssignFunction[f,A,B]";
AssignFunction[__] := $Failed /; Message[AssignFunction::args];
UnorderedPairs::args = "Wrong argument(s), try UnorderedPairs[n]";
UnorderedPairs[__] := $Failed /; Message[UnorderedPairs::args];
PowerSet::args = "Wrong argument(s), try PowerSet[A,k]";
PowerSet[__] := $Failed /; Message[PowerSet::args];
CartesianProduct::args = "Wrong argument(s), try CartesianProduct[A,B]";
CartesianProduct[__] := $Failed /; Message[CartesianProduct::args];
RankCartesian::args = "Wrong argument(s), try RankCartesian[elm,cp]";
RankCartesian[__] := $Failed /; Message[RankCartesian::args];
UnrankCartesian::args = "Wrong argument(s), try UnrankCartesian[r,cp]";
UnrankCartesian[__] := $Failed /; Message[UnrankCartesian::args];
RankingRules::args = "Wrong argument(s), try RankingRules[L]";
RankingRules[__] := $Failed /; Message[RankingRules::args];
Shuffle::args = "Wrong argument(s), try Shuffle[A,B]";
Shuffle[__] := $Failed /; Message[Shuffle::args];
PerfectShuffle::args = "Wrong argument(s), try PerfectShuffle[A,B]";
PerfectShuffle[__] := $Failed /; Message[PerfectShuffle::args];
ToClasses::args = "Wrong argument(s), try ToClasses[L,R]";
ToClasses[__] := $Failed /; Message[ToClasses::args];
EquivalenceToIndex::args = "Wrong argument(s), try EquivalenceToIndex[L,R]";
EquivalenceToIndex[__] := $Failed /; Message[EquivalenceToIndex::args];
AnalyzeOrbit::args = "Wrong argument(s), try AnalyzeOrbit[f,x]";
AnalyzeOrbit[__] := $Failed /; Message[AnalyzeOrbit::args];
CycleDecomposition::args = "Wrong argument(s), try CycleDecomposition[A,f]";
CycleDecomposition[__] := $Failed /; Message[CycleDecomposition::args];
Closure::args = "Wrong argument(s), try Closure[q0,{f,g,..},opts]";
Closure[__] := $Failed /; Message[Closure::args];
ClosureSearch::args = "Wrong argument(s), try ClosureSearch[q0,x,{f,g,..},opts]";
ClosureSearch[__] := $Failed /; Message[ClosureSearch::args];
NonDecreasingSequences::args = "Wrong argument(s), try NonDecreasingSequences[n,k]";
NonDecreasingSequences[__] := $Failed /; Message[NonDecreasingSequences::args];
TransMod::args = "Wrong argument(s), try TransMod[x,t,p]";
TransMod[__] := $Failed /; Message[TransMod::args];



End[];
EndPackage[];
